#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect.ge_discovery_nm670 as gate_spect
from opengate.tests import utility

paths = utility.get_default_test_paths(__file__, "", "test033")


def create_test(sim, nb_thread=1):
    # main options
    sim.g4_verbose = False
    sim.running_verbose_level = gate.logger.RUN
    sim.number_of_threads = nb_thread
    sim.visu = False
    sim.visu_type = "qt"
    sim.visu_verbose = False
    sim.check_volumes_overlap = False
    sim.random_seed = 123456
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    keV = gate.g4_units.keV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq
    sec = gate.g4_units.second
    deg = gate.g4_units.deg
    kBq = 1000 * Bq
    MBq = 1000 * kBq

    ac = 3 * MBq
    distance = 15 * cm
    psd = 6.11 * cm
    p = [0, 0, -(distance + psd)]
    if sim.visu:
        ac = ac / 100

    # world size
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]
    world.material = "G4_AIR"

    # spect head (debug mode = very small collimator)
    spect1, colli, crystal = gate_spect.add_spect_head(
        sim, "spect1", collimator_type="lehr", debug=sim.visu
    )
    spect1.translation, spect1.rotation = gate.geometry.utility.get_transform_orbiting(
        p, "x", 180
    )

    # spect head (debug mode = very small collimator)
    spect2, colli, crystal = gate_spect.add_spect_head(
        sim, "spect2", collimator_type="lehr", debug=sim.visu
    )
    spect2.translation, spect2.rotation = gate.geometry.utility.get_transform_orbiting(
        p, "x", 0
    )

    # physic list
    sim.physics_manager.set_production_cut("world", "all", 10 * mm)
    sim.physics_manager.set_production_cut("spect1_crystal", "all", 1 * mm)
    sim.physics_manager.set_production_cut("spect2_crystal", "all", 1 * mm)

    # source #1
    sources = []
    source = sim.add_source("GenericSource", "source1")
    source.particle = "gamma"
    source.energy.type = "mono"
    source.energy.mono = 140.5 * keV
    source.position.type = "sphere"
    source.position.radius = 2 * mm
    source.position.translation = [0, 0, 20 * mm]
    source.direction.type = "iso"
    source.direction.acceptance_angle.volumes = ["spect1", "spect2"]
    # will be set to false in noaa tests
    source.direction.acceptance_angle.intersection_flag = True
    source.direction.acceptance_angle.normal_flag = True
    source.direction.acceptance_angle.normal_vector = [0, 0, -1]
    source.direction.acceptance_angle.normal_tolerance = 10 * deg
    source.direction.acceptance_angle.skip_policy = "ZeroEnergy"
    source.activity = ac / sim.number_of_threads
    sources.append(source)

    # source #2
    source2 = sim.add_source("GenericSource", "source2")
    # FIXME when source will be refactored, will be possible to use configure_like
    source2.particle = "gamma"
    source2.energy.type = "mono"
    source2.energy.mono = 140.5 * keV
    source2.position.type = "sphere"
    source2.direction.type = "iso"
    source2.direction.acceptance_angle.volumes = ["spect1", "spect2"]
    source2.direction.acceptance_angle.intersection_flag = True
    source2.direction.acceptance_angle.normal_flag = True
    source2.direction.acceptance_angle.normal_vector = [0, 0, -1]
    source2.direction.acceptance_angle.normal_tolerance = 10 * deg
    source2.direction.acceptance_angle.skip_policy = "ZeroEnergy"
    source2.activity = ac / sim.number_of_threads
    source2.position.radius = 1 * mm
    source2.position.translation = [20 * mm, 0, -20 * mm]
    sources.append(source2)

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.output_filename = "test033_stats.txt"

    # add default digitizer (it is easy to change parameters if needed)
    proj = gate_spect.add_simplified_digitizer_tc99m_OLD(
        sim, "spect1_crystal", "test033_proj_1.mhd"
    )
    proj.origin_as_image_center = False
    proj = gate_spect.add_simplified_digitizer_tc99m_OLD(
        sim, "spect2_crystal", "test033_proj_2.mhd"
    )
    proj.origin_as_image_center = False

    # motion of the spect, create also the run time interval
    heads = [spect1, spect2]

    # create a list of run
    n = 9
    sim.run_timing_intervals = gate.runtiming.range_timing(0, 1 * sec, n)
    for head in heads:
        tr, rot = gate.geometry.utility.volume_orbiting_transform(
            "x", 0, 180, n, head.translation, head.rotation
        )
        head.add_dynamic_parametrisation(translation=tr, rotation=rot)

    return sources


def evaluate_test(sim, sources, itol, ref_skipped):
    stats = sim.get_actor("Stats")
    print(stats)
    # ref with _noaa
    # stats.write(paths.output_ref / "test033_stats.txt")

    se = 0
    ze = 0
    for source in sources:
        se += source.total_skipped_events
        ze += source.total_zero_events
    print(f"Skipped particles {se}")
    print(f"Zeros E particles {ze}")
    s = max(se, ze)

    # check nb of avoided events (either skipped or energy zero)
    gate.exception.warning(f"Check nb of skipped particles")
    tol = 0.01
    if ref_skipped != 0:
        d = abs(ref_skipped - s) / ref_skipped
    else:
        d = 0
    is_ok = d < tol
    utility.print_test(
        is_ok,
        f"Skipped particles ref={ref_skipped}, get {s} -> {d * 100}% vs tol={tol * 100}%",
    )

    # check stats
    gate.exception.warning(f"Check stats")
    stats_ref = utility.read_stat_file(paths.output_ref / "test033_stats.txt")
    print(f"Steps counts not compared (was {stats.counts.steps})")
    nbt = sim.number_of_threads
    stats.counts.steps = stats_ref.counts.steps
    stats_ref.counts.runs *= nbt
    if se > 0:
        print(f"Track counts not compared (was {stats.counts.tracks})")
        print(f"Modify Events + skipped {stats.counts.events + se})")
        stats.counts.events += se
        stats.counts.tracks = stats_ref.counts.tracks
    if ze > 0:
        print(f"Track counts not compared (was {stats.counts.tracks})")
        stats.counts.tracks = stats_ref.counts.tracks
    is_ok = utility.assert_stats(stats, stats_ref, 0.03) and is_ok

    # compare edep map
    gate.exception.warning(f"Check images")
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test033_proj_1.mhd",
            paths.output / "test033_proj_1_counts.mhd",
            stats,
            tolerance=68,
            axis="x",
            sum_tolerance=itol,
            ignore_value_data2=0,
            apply_ignore_mask_to_sum_check=False,
        )
        and is_ok
    )
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test033_proj_2.mhd",
            paths.output / "test033_proj_2_counts.mhd",
            stats,
            tolerance=75,
            axis="x",
            sum_tolerance=itol,
            ignore_value_data2=0,
            apply_ignore_mask_to_sum_check=False,  # reproduce legacy behavior of assert_images
        )
        and is_ok
    )

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
from scipy.spatial.transform import Rotation
import gatetools.phsp as phsp
import opengate as gate
from opengate.tests import utility
from pathlib import Path

# units
m = gate.g4_units.m
mm = gate.g4_units.mm
cm = gate.g4_units.cm
nm = gate.g4_units.nm
Bq = gate.g4_units.Bq
MeV = gate.g4_units.MeV
deg = gate.g4_units.deg


def create_test_phs(
    particle="proton",
    phs_name=Path("output") / "test_proton.root",
    number_of_particles=1,
    translation=None,
):
    # create the simulation
    if translation is None:
        translation = [0 * mm, 0 * mm, 0 * mm]
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    # sim.running_verbose_level = gate.EVENT
    sim.number_of_threads = 1
    sim.random_seed = 987654321

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    nm = gate.g4_units.nm
    MeV = gate.g4_units.MeV

    ##########################################################################################
    # geometry
    ##########################################################################################
    #  adapt world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_Galactic"

    # virtual plane for phase space
    plane = sim.add_volume("Tubs", "phase_space_plane")
    # plane.material = "G4_AIR"
    plane.material = "G4_Galactic"
    plane.rmin = 0
    plane.rmax = 30 * cm
    plane.dz = 1 * nm  # half height
    # plane.rotation = Rotation.from_euler("xy", [180, 30], degrees=True).as_matrix()
    # plane.translation = [0 * mm, 0 * mm, 0 * mm]
    plane.translation = translation

    plane.color = [1, 0, 0, 1]  # red

    ##########################################################################################
    # Actors
    ##########################################################################################
    # Split the joined path into the directory path and filename
    directory_path, filename = os.path.split(phs_name)
    # Extract the base filename and extension
    base_filename, extension = os.path.splitext(filename)
    new_extension = ".root"

    # PhaseSpace Actor
    ta1 = sim.add_actor("PhaseSpaceActor", "PhaseSpace1")
    ta1.attached_to = plane.name
    ta1.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "ParticleName",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
    ]
    new_joined_path = os.path.join(directory_path, base_filename + new_extension)
    ta1.output_filename = new_joined_path
    ta1.debug = False
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = particle
    ta1.filters.append(f)

    # PhaseSpace Actor
    ta2 = sim.add_actor("PhaseSpaceActor", "PhaseSpace2")
    ta2.attached_to = plane.name
    ta2.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "PreDirection",
        "PreDirectionLocal",
    ]
    new_joined_path = os.path.join(
        directory_path, base_filename + "_noParticleInfo" + new_extension
    )
    ta2.output_filename = new_joined_path
    ta2.debug = False
    ta2.filters.append(f)

    # PhaseSpace Actor
    ta3 = sim.add_actor("PhaseSpaceActor", "PhaseSpace3")
    ta3.attached_to = plane.name
    ta3.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
    ]
    new_joined_path = os.path.join(
        directory_path, base_filename + "_PDGCode" + new_extension
    )
    ta3.output_filename = new_joined_path
    ta3.debug = False
    ta3.filters.append(f)

    # PhaseSpace Actor
    ta4 = sim.add_actor("PhaseSpaceActor", "PhaseSpace4")
    ta4.attached_to = plane.name
    ta4.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "PreDirection",
        "PreDirectionLocal",
        "ParticleName",
    ]
    new_joined_path = os.path.join(
        directory_path, base_filename + "_ParticleName" + new_extension
    )
    ta4.output_filename = new_joined_path
    ta4.debug = False
    ta4.filters.append(f)

    # sim.physics_manager.physics_list_name = "FTFP_BERT"
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMZ"

    ##########################################################################################
    #  Source
    ##########################################################################################
    source = sim.add_source("GenericSource", "particle_source")
    source.attached_to = "world"
    # source.particle = "ion 6 12"  # Carbon ions
    source.particle = "proton"
    source.position.type = "point"
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.position.translation = [0 * cm, 0 * cm, -0.1 * cm]
    source.energy.type = "mono"
    source.energy.mono = 150 * MeV
    source.n = number_of_particles

    sim.run(start_new_process=True)


def create_phs_without_source(
    phs_name=Path("output") / "test_proton.root",
):
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    # sim.running_verbose_level = gate.EVENT
    sim.number_of_threads = 1
    sim.random_seed = 987654321

    # ui.g4_verbose = True
    # ui.g4_verbose_level = 100

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    nm = gate.g4_units.nm

    ##########################################################################################
    # geometry
    ##########################################################################################
    #  adapt world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_Galactic"
    print(world.name)

    # virtual plane for phase space
    plane = sim.add_volume("Tubs", "phase_space_plane")
    # plane.material = "G4_AIR"
    plane.material = "G4_Galactic"
    plane.rmin = 0
    plane.rmax = 30 * cm
    plane.dz = 1 * nm  # half height
    # plane.rotation = Rotation.from_euler("xy", [180, 30], degrees=True).as_matrix()
    plane.translation = [0 * mm, 0 * mm, 0 * mm]
    # plane.translation = translation

    plane.color = [1, 0, 0, 1]  # red

    ##########################################################################################
    # Actors
    ##########################################################################################

    # PhaseSpace Actor
    ta1 = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    ta1.attached_to = plane.name
    ta1.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "ParticleName",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
    ]
    ta1.output_filename = phs_name
    ta1.steps_to_store = "exiting"
    ta1.debug = True

    # ~ phys.physics_list_name = "FTFP_BERT"
    sim.physics_manager.physics_list_name = "QGSP_BIC_EMZ"

    # ##########################################################################################
    # #  Source
    # ##########################################################################################
    # # phsp source
    # source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    # source.attached_to = world.name
    # source.phsp_file = source_name
    # source.position_key = "PrePosition"
    # source.direction_key = "PreDirection"
    # source.global_flag = True
    # source.particle = particle
    # source.batch_size = 3000
    # source.n = number_of_particles / sim.number_of_threads
    # # source.position.translation = [0 * cm, 0 * cm, -35 * cm]

    # output = sim.run()
    # output = sim.start()
    # output = sim.start(start_new_process=True)
    return sim


def test_source_name(
    source_file_name=Path("output") / "test_proton_offset.root",
    phs_file_name_out=Path("output") / "output/test_source_electron.root",
) -> None:
    sim = create_phs_without_source(
        phs_name=phs_file_name_out,
    )
    particle = "e-"
    number_of_particles = 1
    ##########################################################################################
    #  Source
    ##########################################################################################
    # phsp source
    source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    source.attached_to = "world"
    source.phsp_file = source_file_name
    source.position_key = "PrePositionLocal"
    source.direction_key = "PreDirectionLocal"
    source.global_flag = True
    source.particle = particle
    source.batch_size = 3000
    source.n = number_of_particles
    source.verbose = False
    # source.position.translation = [0 * cm, 0 * cm, -35 * cm]

    sim.run()


def test_source_particle_info_from_phs(
    source_file_name=Path("output") / "test_proton_offset.root",
    phs_file_name_out=Path("output") / "test_source_PDG_proton.root",
) -> None:
    sim = create_phs_without_source(
        phs_name=phs_file_name_out,
    )
    number_of_particles = 1
    ##########################################################################################
    #  Source
    ##########################################################################################
    # phsp source
    source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    source.attached_to = "world"
    source.phsp_file = source_file_name
    source.particle = ""
    source.position_key = "PrePositionLocal"
    source.direction_key = "PreDirectionLocal"
    source.global_flag = True
    source.batch_size = 3000
    source.n = number_of_particles
    # source.position.translation = [0 * cm, 0 * cm, -35 * cm]

    sim.run()


def test_source_translation(
    source_file_name=Path("output") / "test_proton_offset.root",
    phs_file_name_out=Path("output") / "output/test_source_electron.root",
) -> None:
    sim = create_phs_without_source(
        phs_name=phs_file_name_out,
    )
    number_of_particles = 1
    ##########################################################################################
    #  Source
    ##########################################################################################
    # phsp source
    source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    source.attached_to = "world"
    source.phsp_file = source_file_name
    source.position_key = "PrePosition"
    source.direction_key = "PreDirection"
    source.global_flag = True
    source.particle = "proton"
    source.batch_size = 3000
    source.n = number_of_particles
    source.translate_position = True
    source.position.translation = [3 * cm, 0 * cm, 0 * cm]
    print(source)

    sim.run()


def test_source_rotation(
    source_file_name=Path("output") / "test_proton_offset.root",
    phs_file_name_out=Path("output") / "output/test_source_electron.root",
) -> None:
    sim = create_phs_without_source(
        phs_name=phs_file_name_out,
    )
    number_of_particles = 1
    ##########################################################################################
    #  Source
    ##########################################################################################
    # phsp source
    source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    source.attached_to = "world"
    source.phsp_file = source_file_name
    source.position_key = "PrePosition"
    source.direction_key = "PreDirection"
    source.global_flag = True
    source.particle = "proton"
    source.batch_size = 3000
    source.n = number_of_particles
    source.verbose = False
    # source.translate_position = True
    # source.position.translation = [3 * cm, 1 * cm, 0 * cm]
    source.rotate_direction = True
    # rotation = Rotation.from_euler("zyx", [30, 20, 10], degrees=True)
    rotation = Rotation.from_euler("x", [30], degrees=True)
    source.position.rotation = rotation.as_matrix()
    print(source)

    sim.run()


def test_source_until_primary(
    source_file_name=Path("output") / "test_proton_offset.root",
    phs_file_name_out=Path("output") / "output/test_source_electron.root",
) -> None:
    sim = create_phs_without_source(
        phs_name=phs_file_name_out,
    )
    number_of_particles = 2
    # phsp source
    source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    source.attached_to = "world"
    source.phsp_file = source_file_name
    source.position_key = "PrePosition"
    source.direction_key = "PreDirection"
    source.global_flag = True
    source.particle = ""
    source.batch_size = 3000
    source.n = number_of_particles
    source.generate_until_next_primary = True
    source.primary_lower_energy_threshold = 90.0 * MeV
    source.primary_PDGCode = 2212
    print(source)

    sim.run()


def get_first_entry_of_key(
    file_name_root=Path("output") / "test_source_electron.root", key="ParticleName"
) -> None:
    # read root file
    data_ref, keys_ref, m_ref = phsp.load(str(file_name_root))
    # print(data_ref)
    # print(keys_ref)
    index = keys_ref.index(key)
    # print(index)
    # print(data_ref[index][0])
    return data_ref[0][index]


def check_value_from_root_file(
    file_name_root=Path("output") / "test_source_electron.root",
    key="ParticleName",
    ref_value="e-",
):
    # read root file
    value = get_first_entry_of_key(file_name_root=file_name_root, key=key)
    if (type(ref_value) != str) and (type(value) != str):
        is_ok = utility.check_diff_abs(
            float(value), float(ref_value), tolerance=1e-6, txt=key
        )
    # utility.check_diff_abs(float(value), float(ref_value), tolerance=1e-6, txt=key)
    else:
        if value == ref_value:
            # print("Is correct")
            is_ok = True
        else:
            # print("Is not correct")
            is_ok = False
    # print("ref_value: ", ref_value)
    # print("value: ", value)
    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from test053_phid_helpers2 import *
import opengate as gate


def main(dependency="test053_phid_05_it_ref_mt.py"):
    paths = get_default_test_paths(__file__, "", output_folder="test053")

    # bi213 83 213
    # ac225 89 225
    # fr221 87 221
    z = 89
    a = 225
    nuclide, _ = get_nuclide_and_direct_progeny(z, a)
    print(nuclide)

    # this test need the test053_phid_05 before
    root_ref = paths.output / f"test053_{nuclide.nuclide}_5_ref.root"
    print(root_ref)
    if not os.path.exists(root_ref):
        # ignore on windows
        if os.name == "nt":
            test_ok(True)
            sys.exit(0)
        cmd = "python " + str(paths.current / dependency)
        print(cmd)
        r = os.system(cmd)

    sim = gate.Simulation()
    sim_name = f"{nuclide.nuclide}_6_model"
    create_sim_test053(sim, sim_name)

    # sources
    activity_in_bq = 1000
    s = add_source_model(sim, z, a, activity_in_bq)
    s.atomic_relaxation_flag = False
    s.isomeric_transition_flag = True

    # go
    sec = g4_units.second
    minute = g4_units.minute
    start_time = 19 * minute
    end_time = start_time + 10 * sec
    duration = end_time - start_time
    print(f"start time {start_time / sec}")
    print(f"end time {end_time / sec}")
    print(f"Duration {duration / sec}")
    print(f"Ions {activity_in_bq * duration / sec:.0f}")
    sim.run_timing_intervals = [[start_time, end_time]]

    # go
    sim.run(start_new_process=True)

    # print stats
    stats = sim.get_actor("stats")
    print(stats)

    # compare
    print()
    print(f"check root files")
    root_ref = paths.output / f"test053_{nuclide.nuclide}_5_ref.root"
    root_model = sim.get_actor("phsp").get_output_path()
    print(f"{root_ref = }")
    print(f"{root_model = }")
    is_ok = compare_root_energy(
        root_ref, root_model, start_time, end_time, model_index=129, tol=0.09
    )
    test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.tests.utility as utility
from opengate.utility import g4_units

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test004_simulation_stats_actor"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = True
    sim.visu_type = "qt"
    sim.visu_verbose = True
    sim.number_of_threads = 1
    sim.random_engine = "MersenneTwister"
    sim.random_seed = "auto"

    # set the world size like in the Gate macro
    m = g4_units.m
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]

    # add a simple waterbox volume
    waterbox = sim.add_volume("Box", "Waterbox")
    cm = g4_units.cm
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    # default source for tests
    keV = g4_units.keV
    mm = g4_units.mm
    Bq = g4_units.Bq
    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 200 * Bq

    # runs
    sec = g4_units.second
    sim.run_timing_intervals = [[0, 0.5 * sec], [0.5 * sec, 1.0 * sec]]

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")

    # start simulation
    # sim.g4_commands_after_init.append("/run/verbose 1")
    sim.run()

    print(stats)
    utility.test_ok(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import subprocess
from matplotlib import pyplot as plt
from opengate.tests import utility
from test085_free_flight_helpers import *
import uproot
import os, sys


def main(dependency="test085_free_flight_phsp_1_ref_mt.py"):

    paths = utility.get_default_test_paths(__file__, None, output_folder="test085_phsp")

    # not on windows
    if os.name == "nt":
        sys.exit(0)

    # The test needs the output of the other tests
    if not os.path.isfile(paths.output / "phsp_sphere_ref.root"):
        subprocess.call(["python", paths.current / dependency])
    if not os.path.isfile(paths.output / "phsp_sphere_ff.root"):
        subprocess.call(["python", paths.current / "test085_free_flight_phsp_2_ff.py"])
    if not os.path.isfile(paths.output / "phsp_sphere_ff_sc.root"):
        subprocess.call(
            ["python", paths.current / "test085_free_flight_phsp_3_scatter_mt.py"]
        )

    # read root output
    ref_filename = paths.output / "phsp_sphere_ref.root"
    prim_filename = paths.output / "phsp_sphere_ff.root"
    sca_filename = paths.output / "phsp_sphere_ff_sc.root"
    ref_n = 5e4
    prim_n = 1e4
    sca_n = 2e3
    scaling_sc = ref_n / sca_n
    scaling_prim = ref_n / prim_n

    # energy histo
    branch = "phsp_sphere"
    k = "KineticEnergy"
    ene_ref = uproot.open(paths.output / ref_filename)[branch]
    ene_ref = ene_ref.arrays(library="numpy")[k]

    ene_prim = uproot.open(paths.output / prim_filename)[branch]
    ene_prim_w = ene_prim.arrays(library="numpy")["Weight"] * scaling_prim
    ene_prim = ene_prim.arrays(library="numpy")[k]

    ene_sc = uproot.open(paths.output / sca_filename)[branch]
    ene_sc_w = ene_sc.arrays(library="numpy")["Weight"] * scaling_sc
    ene_sc = ene_sc.arrays(library="numpy")[k]

    print(
        f"Number of events, ref    = {len(ene_ref)}  {np.min(ene_ref):.2f} {np.max(ene_ref):.2f}"
    )
    print(
        f"Number of events, prim   = {len(ene_prim)} {np.min(ene_prim):.2f} {np.max(ene_prim):.2f}"
    )
    print(
        f"Number of events, sc     = {len(ene_sc)} {np.min(ene_sc):.2f} {np.max(ene_sc):.2f}"
    )
    print(f"scaling_prim             = {scaling_prim}")
    print(f"scaling_sc               = {scaling_sc}")

    print()
    print("==== TOTAL")
    print(f"Sum of weights, prim     = {ene_prim_w.sum():.0f}")
    print(f"Sum of weights, sec      = {ene_sc_w.sum():.0f}")
    t = ene_sc_w.sum() + ene_prim_w.sum()
    print(f"Sum of weights, p+sec    = {t:.0f} vs {len(ene_ref)}")

    check1 = (len(ene_ref) - (ene_sc_w.sum() + ene_prim_w.sum())) / len(ene_ref) * 100
    tol = 1.0
    is_ok = np.fabs(check1) < tol
    utility.print_test(
        is_ok, f"diff p+sec               = {check1:.2f} %     tol={tol:.2f}"
    )
    print(
        f"total rel diff ref-prim  = "
        f"{(len(ene_ref) - ene_prim_w.sum())/len(ene_ref)*100:.2f} %"
    )
    print(
        f"total rel diff ref-sec   = "
        f"{(len(ene_ref) - ene_sc_w.sum())/len(ene_ref)*100:.2f} %"
    )

    print()
    e_threshold = 0.14051  # 0.140511
    ene_ref_peak = ene_ref[ene_ref >= e_threshold]
    ene_sc_peak_w = ene_sc_w[ene_sc >= e_threshold]
    print("==== PEAK")
    print(f"Number of peaks ref      = {len(ene_ref_peak)}")
    print(f"Number of peaks prim     = {ene_prim_w.sum():.0f}")
    print(f"Number of peaks sec      = {ene_sc_peak_w.sum():.0f}")
    d = len(ene_ref_peak) - ene_prim_w.sum()
    check2 = d / len(ene_ref_peak) * 100
    tol = 10.0
    b = np.fabs(check2) < tol and np.fabs(check2) > 2
    utility.print_test(
        b,
        f"diff peak ref-prim       = {check2:.2f} %  (diff={d:.0f}) tol={tol} and >2%",
    )
    is_ok = b and is_ok

    # add the peak from the scatter (should be low)
    d = len(ene_ref_peak) - (ene_prim_w.sum() + ene_sc_peak_w.sum())
    check3 = d / len(ene_ref_peak) * 100
    tol = 1.0
    b = np.fabs(check3) < tol
    utility.print_test(b, f"diff peaks ref-(p+sc)    = {check3:.2f} %    tol={tol}")
    is_ok = b and is_ok

    # scatter
    print()
    print("==== SCATTER")
    ene_ref_scatter = ene_ref[ene_ref < e_threshold]
    ene_prim_w_scatter = ene_prim_w[ene_prim < e_threshold]  # should be almost zero
    ene_sc_w_scatter = ene_sc_w[ene_sc < e_threshold]
    print(f"Number of scatter ref      = {len(ene_ref_scatter)}")
    print(f"Number of scatter prim     = {ene_prim_w_scatter.sum():.0f}")
    print(f"Number of scatter sec      = {ene_sc_w_scatter.sum():.0f}")
    d = len(ene_ref_scatter) - (ene_sc_w_scatter.sum() + ene_prim_w_scatter.sum())
    check4 = d / len(ene_ref_scatter) * 100
    tol = 1.0
    b = np.fabs(check4) < tol
    utility.print_test(b, f"diff scatter ref-sc        = {check4:.2f} %    tol={tol}")
    is_ok = b and is_ok

    # Key KineticEnergy min/mean/max: 0.08349965517188826 0.1404070326435755 0.140511

    ene_total = np.concatenate((ene_prim, ene_sc))
    w_total = np.concatenate((ene_prim_w, ene_sc_w))

    fig = plt.figure()
    ax = fig.add_subplot(111)
    bins = 100
    d = False
    r = (0.0, 0.145)
    ax.hist(ene_ref, bins=bins, alpha=0.5, density=d, label="ref", range=r)
    ax.hist(
        ene_prim,
        bins=bins,
        alpha=0.5,
        density=d,
        range=r,
        weights=ene_prim_w,
        label="prim",
    )
    ax.hist(
        ene_sc, bins=bins, range=r, weights=ene_sc_w, alpha=0.5, density=d, label="sc"
    )

    ax.hist(
        ene_total,
        bins=bins,
        range=r,
        weights=w_total,
        alpha=0.2,
        density=d,
        label="total",
        histtype="stepfilled",
    )

    f = paths.output / "phsp_total.png"
    plt.legend()
    plt.savefig(f)
    print(f)
    # plt.show()

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from opengate.devtools import find_unprocessed_gateobject_classes
from opengate.exception import GateImplementationError
import opengate.tests.utility as utility


if __name__ == "__main__":
    is_ok = True
    exceptions = []

    unprocessed_classes = find_unprocessed_gateobject_classes()
    if len(unprocessed_classes) > 0:
        is_ok = False
        s = "\n".join([f"{i}) {w}" for i, w in enumerate(unprocessed_classes)])
        exceptions.append(
            GateImplementationError(
                f"{len(unprocessed_classes)} GateObjects are not processed upon import: \n{s}"
            )
        )
    else:
        print("All classes inheriting from GateObject are properly processed .")
    utility.test_ok(is_ok, exceptions=exceptions)



import numpy as np
import pandas as pd
import os
from scipy.spatial.transform import Rotation
from pathlib import Path
import SimpleITK as sitk

import opengate_core as g4
from .base import ActorBase
from ..exception import fatal, warning
from ..utility import (
    g4_units,
    standard_error_c4_correction,
)
from ..image import (
    update_image_py_to_cpp,
    get_py_image_from_cpp_image,
    images_have_same_domain,
    resample_itk_image_like,
    itk_image_from_array,
    divide_itk_images,
    scale_itk_image,
)
from ..geometry.utility import get_transform_world_to_local
from ..geometry.materials import create_density_img
from ..base import process_cls
from .actoroutput import (
    ActorOutputSingleImage,
    ActorOutputSingleMeanImage,
    ActorOutputQuotientMeanImage,
    ActorOutputSingleImageWithVariance,
    UserInterfaceToActorOutputImage,
)
from .dataitems import (
    ItkImageDataItem,
)


class VoxelDepositActor(ActorBase):
    """Base class which holds user input parameters common to all actors
    that deposit quantities in a voxel grid, e.g. the DoseActor.
    """

    # hints for IDE
    size: list
    spacing: list
    translation: list
    rotation: list
    repeated_volume_index: int
    hit_type: str
    output: str
    img_coord_system: str
    output_coordinate_system: str

    user_info_defaults = {
        "size": (
            [10, 10, 10],
            {
                "doc": "Size of the dose grid in number of voxels [N_x, N_y, N_z]. Expects a list of integer values of size 3.",
            },
        ),
        "spacing": (
            [1 * g4_units.mm, 1 * g4_units.mm, 1 * g4_units.mm],
            {
                "doc": "Voxel spacing vector along the [x, y, z] coordinates with units. "
                "The user sets the units by multiplication with g4_units.XX. "
                "The default unit is g4_unit.mm amd the default spacing is [1, 1, 1] [g4_units.mm]",
            },
        ),
        "translation": (
            [0 * g4_units.mm, 0 * g4_units.mm, 0 * g4_units.mm],
            {
                "doc": "Translation vector to (optionally) translate the image in along [x, y, z] from the center of the attached volume. The default unit is g4_units.mm and default value is the unity operation [0, 0, 0] * g4_units.mm. ",
            },
        ),
        "rotation": (
            Rotation.identity().as_matrix(),
            {
                "doc": "Rotation matrix to (optionally) rotate the image. Default is the identiy matrix.",
            },
        ),
        "repeated_volume_index": (
            0,
            {
                "doc": "Index of the repeated volume (G4PhysicalVolume) to which this actor is attached. "
                "For non-repeated volumes, this value is always 0. ",
            },
        ),
        "hit_type": (
            "random",
            {
                "doc": "For advanced users: define the position of interaction to which the deposited quantity is associated to, "
                "i.e. at the Geant4 PreStepPoint, PostStepPoint, or somewhere in between (middle or (uniform) random). In doubt use/start with random.",
                "allowed_values": ("random", "pre", "post", "middle"),
            },
        ),
        "output": (
            None,
            {
                "deprecated": "The output filename is now set via output_filename relative to the output "
                "directory of the simulation, which can be set via sim.output_dir. "
                "If no output_filename is provided, it will be generated automatically. "
                "To specify whether the actor output should be written to disk, use write_to_disk=True/False."
            },
        ),
        "img_coord_system": (
            None,
            {
                "deprecated": "The user input parameter 'img_coord_system' is deprecated. "
                "Use my_actor.output_coordinate_system='attached_to_image' instead, "
                "where my_actor should be replaced with your actor object. ",
            },
        ),
        "output_coordinate_system": (
            "local",
            {
                "doc": "This command sets the reference coordinate system, which can be the local volume (attached_to commmand), global or attached to image.",
                "allowed_values": ("local", "global", "attached_to_image", None),
            },
        ),
    }

    def __init__(self, *args, **kwargs) -> None:
        ActorBase.__init__(self, *args, **kwargs)

    def check_user_input(self):
        if self.output_coordinate_system == "attached_to_image":
            if not hasattr(
                self.attached_to_volume, "native_translation"
            ) or not hasattr(self.attached_to_volume, "native_rotation"):
                fatal(
                    f"User input 'output_coordinate_system' = {self.output_coordinate_system} "
                    f"of actor {self.name} is not compatible "
                    f"with the volume to which this actor is attached: "
                    f"{self.attached_to} ({self.attached_to_volume.volume_type})"
                )

    def initialize(self):
        super().initialize()

        msg = (
            f"cannot be used in actor {self.name} "
            f"because the volume ({self.attached_to}, {self.attached_to_volume.type_name}) "
            f"to which the actor is attached does not support it. "
        )
        if isinstance(self.spacing, str) and self.spacing == "like_image_volume":
            if not hasattr(self.attached_to_volume, "spacing"):
                fatal("spacing = 'like_image_volume' " + msg)
            self.spacing = self.attached_to_volume.spacing
        if isinstance(self.size, str) and self.size == "like_image_volume":
            if not hasattr(self.attached_to_volume, "size_pix"):
                fatal("size = 'like_image_volume' " + msg)
            self.size = self.attached_to_volume.size_pix

    def get_physical_volume_name(self):
        # init the origin and direction according to the physical volume
        # (will be updated in the BeginOfRun)
        if self.repeated_volume_index is None:
            repeated_volume_index = 0
        else:
            repeated_volume_index = self.repeated_volume_index
        try:
            g4_phys_volume = self.attached_to_volume.g4_physical_volumes[
                repeated_volume_index
            ]
        except IndexError:
            fatal(
                f"Error in the actor {self.type_name} named {self.name}. "
                f"Could not find the physical volume with index {repeated_volume_index} "
                f"in volume '{self.attached_to}' to which this actor is attached. "
            )
        # Return the real physical volume name
        return str(g4_phys_volume.GetName())

    def _update_output_coordinate_system(self, which_output, run_index):
        """Method to be called at the end of a run.
        Note: The output image is aligned with the volume to which the actor as attached
        at the beginning of a run. So for the option output_coordinate_system='global',
        nothing has to be done here.
        """
        size = np.array(self.size)
        spacing = np.array(self.spacing)
        origin = -size * spacing / 2.0 + spacing / 2.0

        translation = np.array(self.translation)
        origin_local = Rotation.from_matrix(self.rotation).apply(origin) + translation

        # image centered at (0,0,0), no rotation
        if self.output_coordinate_system is None:
            self.user_output[which_output].set_image_properties(
                run_index,
                origin=origin.tolist(),
                rotation=Rotation.identity().as_matrix(),
            )
        # image centered at self.translation and rotated by self.rotation,
        # i.e. in the reference frame of the volume to which the actor is attached.
        elif self.output_coordinate_system in ("local",):
            self.user_output[which_output].set_image_properties(
                run_index, origin=origin_local.tolist(), rotation=self.rotation
            )
        # only applicable if the attached_to volume is an image volume
        # as 'local', but considering the origin and direction (rotation) of the image
        # used to create the image volume. Useful overlay output and image volume for further analysis
        elif self.output_coordinate_system in ("attached_to_image",):
            native_origin_image = self.attached_to_volume.native_translation
            native_rotation_image = self.attached_to_volume.native_rotation
            origin_wrt_image = (
                Rotation.from_matrix(native_rotation_image).apply(origin_local)
                + native_origin_image
            )
            rotation_wrt_image = np.matmul(native_rotation_image, self.rotation)

            self.user_output[which_output].set_image_properties(
                run_index, origin=origin_wrt_image.tolist(), rotation=rotation_wrt_image
            )
        elif self.output_coordinate_system in ("global",):
            translation_phys_vol, rotation_phys_vol = get_transform_world_to_local(
                self.attached_to_volume, self.repeated_volume_index
            )
            origin_global = (
                Rotation.from_matrix(rotation_phys_vol).apply(origin_local)
                + translation_phys_vol
            )
            rotation_global = np.matmul(rotation_phys_vol, self.rotation)
            self.user_output[which_output].set_image_properties(
                run_index, origin=origin_global.tolist(), rotation=rotation_global
            )
        else:
            fatal(
                f"Illegal parameter 'output_coordinate_system': {self.output_coordinate_system}"
            )

    def prepare_output_for_run(self, output_name, run_index, **kwargs):
        self._assert_output_exists(output_name)
        # self.user_output[output_name].size = self.size
        # self.user_output[output_name].spacing = self.spacing
        self.user_output[output_name].create_empty_image(
            run_index, self.size, self.spacing, origin=self.translation, **kwargs
        )

    def fetch_from_cpp_image(self, output_name, run_index, *cpp_image):
        self._assert_output_exists(output_name)
        data = []
        for i, cppi in enumerate(cpp_image):
            if self.user_output[output_name].get_active(item=i):
                py_image = get_py_image_from_cpp_image(cppi, view=False)
                # FIXME: not needed, I think, because get_py_image_from_cpp_image copies spacing and origin
                # There is an empty image already which has served as storage for meta info like size and spacing.
                # So we get this info back
                # py_image.CopyInformation(
                #     self.user_output[output_name].get_data(run_index, item=i)
                # )
                data.append(py_image)
            else:
                data.append(None)
        self.user_output[output_name].store_data(run_index, *data)

    def push_to_cpp_image(self, output_name, run_index, *cpp_image, copy_data=True):
        self._assert_output_exists(output_name)
        for i, cppi in enumerate(cpp_image):
            if self.user_output[output_name].get_active(item=i):
                update_image_py_to_cpp(
                    self.user_output[output_name].get_data(run_index, item=i),
                    cppi,
                    copy_data,
                )

    def EndOfRunActionMasterThread(self, run_index):
        # inform actor output that this run is over
        for u in self.user_output.values():
            if u.get_active(item="all"):
                u.end_of_run(run_index)
        return 0

    def StartSimulationAction(self):
        # inform actor output that this simulation is starting
        for u in self.user_output.values():
            if u.get_active(item="any"):
                u.start_of_simulation()

    def EndSimulationAction(self):
        # inform actor output that this simulation is over and write data
        for u in self.user_output.values():
            if u.get_active(item="any"):
                u.end_of_simulation()


def _setter_hook_ste_of_mean_unbiased(self, value):
    if value is True:
        self.ste_of_mean = True
    return value


def _setter_hook_ste_of_mean(self, value):
    if value is True:
        self.square = True
        self.use_more_ram = True
    return value


def _setter_hook_uncertainty(self, value):
    if value is True:
        self.square = True
    return value


def _setter_hook_uncertainty_goal(self, value):
    if value is not None and (value < 0.0 or value > 1.0):
        fatal(
            f"Uncertainty goal must be > 0 and < 1, where 1 means 100%. The provided value is: {value}"
        )
    return value


class DoseActor(VoxelDepositActor, g4.GateDoseActor):
    """Computes the deposited energy or absorbed dose in the volume to which it is attached.
    It creates a virtual voxelized scoring image, whose shape and position can be defined by the user.
    """

    # hints for IDE
    score_in: str

    user_info_defaults = {
        "square": (
            True,
            {
                "doc": "This option will provide an additional output image with the squared energy (or dose) deposited per event. This image can be used to calculate the variance of the output variable, as Var(X) = E[X^2] - E[X]^2. This option enables the E[X^2] image.",
                "deprecated": "Use: my_actor.square.active=True/False "
                "to request uncertainty scoring of the respective quantity, "
                "where 'my_actor' should be your actor object. "
                "Note: activating user_output.edep_uncertainty or user_output.dose_uncertainty "
                "implies activating user_output.square. ",
            },
        ),
        "uncertainty": (
            True,
            {
                "doc": "This option will create an additional output image providing the uncertainty of the scored variable (dose or edep).",
                "deprecated": "Use: my_actor.dose_uncertainty.active=True/False and"
                "my_actor.user_output.edep_uncertainty.active=True/False "
                "to request uncertainty scoring of the respective quantity, "
                "where 'my_actor' should be your actor object. ",
                # "setter_hook": _setter_hook_uncertainty,
            },
        ),
        "dose": (
            False,
            {
                "doc": "This option will enable the calculation of the dose image.",
                "deprecated": "Use: my_actor.dose.active=True/False "
                "to request the actor to score dose, "
                "where 'my_actor' should be your actor object. "
                "By default, only the deposited energy is scored. ",
            },
        ),
        "to_water": (
            False,
            {
                "doc": "This option will convert the dose image to dose to water.",
                "deprecated": "Use my_dose_actor.score_in='G4_WATER' instead. ",
            },
        ),
        "score_in": (
            "material",
            {
                "doc": """The score_in command allows to convert the LET from the material, which is defined in the geometry, to any user defined material. Note, that this does not change the material definition in the geometry. The default value is 'material', which means that no conversion is performed and the LET to the local material is scored. You can use any material defined in the simulation or pre-defined by Geant4 such as 'G4_WATER', which may be one of the most use cases of this functionality.
                """,
                "allowed_values": (
                    "material",
                    "G4_WATER",
                ),
            },
        ),
        "ste_of_mean": (
            False,
            {
                "doc": "Calculate the standard error of the mean. Only working in MT mode and the number of threads are considered the sample. To have a meaningful uncertainty at least 8 threads are needed.",
                "setter_hook": _setter_hook_ste_of_mean,
                "deactivated": True,
            },
        ),
        "ste_of_mean_unbiased": (
            False,
            {
                "doc": "Similar to ste_of_mean, but compensates for a bias in ste_of_mean for small sample sizes (<8).  ",
                "setter_hook": _setter_hook_ste_of_mean_unbiased,
                "deactivated": True,
            },
        ),
        "uncertainty_goal": (
            None,
            {
                "doc": "If set, it defines the statistical uncertainty goal. The simulation will stop once the statistical uncertainty is smaller or equal this value.",
                "setter_hook": _setter_hook_uncertainty_goal,
            },
        ),
        "uncertainty_first_check_after_n_events": (
            1e4,
            {
                "doc": "Only applies if uncertainty_goal is set True: Number of events after which uncertainty is evaluated the first time. "
                "After the first evaluation, the value is updated with an estimation of the N events needed to achieve the uncertainty goal, Therefore it is recommended to select a sufficiently large number so the uncertainty of the uncertainty is not too large.",
            },
        ),
        "uncertainty_voxel_edep_threshold": (
            0.7,
            {
                "doc": "Only applies if uncertainty_goal is set True: The calculation of the mean uncertainty of the edep image, only voxels that are above this relative threshold are considered. The threshold must range between [0, 1] and gives the fraction relative to max edep value in the image.",
            },
        ),
        "uncertainty_overshoot_factor_N_events": (
            1.05,
            {
                "doc": "Only applies if uncertainty_goal is set True: Factor multiplying the estimated N events needed to achieve the uncertainty goal, to ensure convergence.",
            },
        ),
        "dose_calc_on_the_fly": (
            False,
            {
                "doc": "FIXME",
                "deactivated": True,
            },
        ),
    }

    user_output_config = {
        "edep_with_uncertainty": {
            "actor_output_class": ActorOutputSingleImageWithVariance,
            "interfaces": {
                "edep": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 0,
                    "active": True,
                },
                "edep_squared": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 1,
                    "active": False,
                },
                "edep_uncertainty": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": "uncertainty",
                    "active": False,
                },
            },
        },
        "dose_with_uncertainty": {
            "actor_output_class": ActorOutputSingleImageWithVariance,
            "interfaces": {
                "dose": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 0,
                    "active": False,
                },
                "dose_squared": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 1,
                    "active": False,
                },
                "dose_uncertainty": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": "uncertainty",
                    "active": False,
                },
            },
        },
        "density": {
            "actor_output_class": ActorOutputSingleMeanImage,
            "active": False,
        },
        "counts": {
            "actor_output_class": ActorOutputSingleImage,
            "active": False,
        },
    }

    def __init__(self, *args, **kwargs):
        VoxelDepositActor.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateDoseActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "BeginOfRunActionMasterThread",
                "EndOfRunActionMasterThread",
                "BeginOfRunAction",
                "EndOfRunAction",
                "BeginOfEventAction",
                "SteppingAction",
                "EndOfEventAction",
            }
        )

    def create_density_image_from_image_volume(self, deposit_image):
        if self.attached_to_volume.volume_type != "ImageVolume":
            fatal(
                f"Cannot calculate the density map from the ImageVolume "
                f"because this actor is attached to a {self.attached_to_volume.volume_type}. "
            )

        density_image = self.attached_to_volume.create_density_image()
        if images_have_same_domain(deposit_image, density_image) is False:
            density_image = resample_itk_image_like(
                density_image, deposit_image, 0, linear=True
            )
        return density_image

    def initialize(self, *args):
        """
        At the start of the run, the image is centered according to the coordinate system of
        the attached volume. This function computes the correct origin = center + translation.
        Note that there is a half-pixel shift to align according to the center of the pixel,
        like in ITK.
        """
        self.check_user_input()

        VoxelDepositActor.initialize(self)

        # the edep component has to be active in any case
        self.user_output.edep_with_uncertainty.set_active(True, item=0)

        # Make sure the squared component (item 1) is active if any of the quantities relying on it are active
        if (
            self.user_output.edep_with_uncertainty.get_active(
                item=("uncertainty", "std", "variance")
            )
            is True
            or self.uncertainty_goal is not None
        ):
            # activate the squared component, but avoid writing it to disk
            # because the user has not activated it and thus most likely does not want it
            if not self.user_output.edep_with_uncertainty.get_active(item=1):
                self.user_output.edep_with_uncertainty.set_write_to_disk(False, item=1)
                self.user_output.edep_with_uncertainty.set_active(
                    True, item=1
                )  # activate squared component

        # Make sure the squared component (item 1) is active if any of the quantities relying on it are active
        if (
            self.user_output.dose_with_uncertainty.get_active(
                item=("uncertainty", "std", "variance")
            )
            is True
        ):
            # activate the dose component
            self.user_output.dose_with_uncertainty.set_active(True, item=0)
            # activate the squared component, but avoid writing it to disk
            # because the user has not activated it and thus most likely does not want it
            if not self.user_output.dose_with_uncertainty.get_active(item=1):
                self.user_output.dose_with_uncertainty.set_write_to_disk(False, item=1)
                self.user_output.dose_with_uncertainty.set_active(
                    True, item=1
                )  # activate squared component

        if (
            self.user_output.density.get_active() is True
            and self.attached_to_volume.volume_type != "ImageVolume"
        ):
            fatal(
                "The dose actor can only produce a density map if it is attached to an ImageVolume. "
                f"This actor is attached to a {self.attached_to_volume.volume_type} volume. "
            )

        self.InitializeUserInfo(self.user_info)  # C++ side
        # Set the flags on C++ side so the C++ knows which quantities need to be scored
        self.SetEdepSquaredFlag(
            self.user_output.edep_with_uncertainty.get_active(item=1)
        )
        self.SetDoseFlag(self.user_output.dose_with_uncertainty.get_active(item=0))
        self.SetDoseSquaredFlag(
            self.user_output.dose_with_uncertainty.get_active(item=1)
        )
        # item=0 is the default
        self.SetCountsFlag(self.user_output.counts.get_active())
        # C++ side has a boolean toWaterFlag and self.score_in == "water" yields True/False
        self.SetToWaterFlag(self.score_in == "G4_WATER")

        # variables for stop on uncertainty functionality
        if self.uncertainty_goal is None:
            self.SetUncertaintyGoal(0)
        else:
            self.SetUncertaintyGoal(self.uncertainty_goal)
        self.SetThreshEdepPerc(self.uncertainty_voxel_edep_threshold)
        self.SetOvershoot(self.uncertainty_overshoot_factor_N_events)
        self.SetNbEventsFirstCheck(int(self.uncertainty_first_check_after_n_events))

        # Set the physical volume name on the C++ side
        self.SetPhysicalVolumeName(self.get_physical_volume_name())
        self.InitializeCpp()

    def BeginOfRunActionMasterThread(self, run_index):
        self.prepare_output_for_run("edep_with_uncertainty", run_index)
        self.push_to_cpp_image(
            "edep_with_uncertainty",
            run_index,
            self.cpp_edep_image,
            self.cpp_edep_squared_image,
        )

        if self.user_output.dose_with_uncertainty.get_active(item="any"):
            self.prepare_output_for_run("dose_with_uncertainty", run_index)
            self.push_to_cpp_image(
                "dose_with_uncertainty",
                run_index,
                self.cpp_dose_image,
                self.cpp_dose_squared_image,
            )

        if self.user_output.counts.get_active():
            self.prepare_output_for_run("counts", run_index)
            self.push_to_cpp_image("counts", run_index, self.cpp_counts_image)

        g4.GateDoseActor.BeginOfRunActionMasterThread(self, run_index)

    def EndOfRunActionMasterThread(self, run_index):
        self.fetch_from_cpp_image(
            "edep_with_uncertainty",
            run_index,
            self.cpp_edep_image,
            self.cpp_edep_squared_image,
        )
        self._update_output_coordinate_system("edep_with_uncertainty", run_index)
        self.user_output.edep_with_uncertainty.store_meta_data(
            run_index, number_of_samples=self.NbOfEvent
        )

        if self.user_output.dose_with_uncertainty.get_active(item="any"):
            self.fetch_from_cpp_image(
                "dose_with_uncertainty",
                run_index,
                self.cpp_dose_image,
                self.cpp_dose_squared_image,
            )
            self._update_output_coordinate_system("dose_with_uncertainty", run_index)
            self.user_output.dose_with_uncertainty.store_meta_data(
                run_index, number_of_samples=self.NbOfEvent
            )
            # divide by voxel volume and scale to unit Gy
            if self.user_output.dose_with_uncertainty.get_active(item=0):
                self.user_output.dose_with_uncertainty.data_per_run[run_index].data[
                    0
                ] /= (g4_units.Gy * self.spacing[0] * self.spacing[1] * self.spacing[2])
            if self.user_output.dose_with_uncertainty.get_active(item=1):
                # in the squared component 1, the denominator needs to be squared
                self.user_output.dose_with_uncertainty.data_per_run[run_index].data[
                    1
                ] /= (
                    g4_units.Gy * self.spacing[0] * self.spacing[1] * self.spacing[2]
                ) ** 2

        if self.user_output.counts.get_active():
            self.fetch_from_cpp_image("counts", run_index, self.cpp_counts_image)
            self._update_output_coordinate_system("counts", run_index)
            self.user_output.counts.store_meta_data(
                run_index, number_of_samples=self.NbOfEvent
            )

        # density image
        if self.user_output.density.get_active():
            edep_image = self.user_output.edep_with_uncertainty.get_data(
                run_index, item=0
            )
            self.user_output.density.store_data(
                run_index, self.create_density_image_from_image_volume(edep_image)
            )
            self.user_output.density.store_meta_data(
                run_index, number_of_samples=self.NbOfEvent
            )

        VoxelDepositActor.EndOfRunActionMasterThread(self, run_index)

        # FIXME: should check if uncertainty goal is reached (return value: 0),
        # but the current mechanism is quite hacky and it is therefore temporarily not in use!
        return 0

    def EndSimulationAction(self):
        g4.GateDoseActor.EndSimulationAction(self)
        VoxelDepositActor.EndSimulationAction(self)


class TLEDoseActor(DoseActor, g4.GateTLEDoseActor):
    """TLE = Track Length Estimator"""

    energy_min: float
    range_type: str
    max_range: float
    database: str

    user_info_defaults = {
        "energy_min": (
            0.0,
            {"doc": "Kill the gamma if below this energy"},
        ),
        "tle_threshold": (
            np.inf,
            {
                "doc": "Define a criterium to enable TLE or not. It can be in terms of gamma energy or in secondary particle range depending on the provided tle_threshold_type"
            },
        ),
        "tle_threshold_type": (
            "None",
            {
                "doc": "Define the type of range lim provided to hTLE. It could be applied without threshold (None), by energy (energy), or by the range of an electron with the full gamma energy (max range) or the average transfered energy (average range)."
            },
        ),
        "database": (
            "EPDL",
            {
                "doc": "which database to use",
                "allowed_values": ("EPDL", "NIST"),  # "simulated" does not work
            },
        ),
    }

    def __initcpp__(self):
        g4.GateTLEDoseActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "BeginOfRunAction",
                "BeginOfEventAction",
                "SteppingAction",
                "PreUserTrackingAction",
                "EndOfRunAction",
            }
        )

    def initialize(self, *args):
        if self.score_in != "material":
            fatal(
                f"TLEDoseActor cannot score in {self.score_in}, only 'material' is allowed."
            )
        super().initialize(args)


def _setter_hook_score_in_let_actor(self, value):
    if value.lower() in ("g4_water", "g4water"):
        """Assuming a misspelling of G4_WATER and correcting it to correct spelling; Note that this is rather dangerous operation."""
        return "G4_WATER"
    else:
        return value


class LETActor(VoxelDepositActor, g4.GateLETActor):
    """This actor scores the Linear Energy Transfer (LET) on a voxel grid in the volume to which the actor is attached. Note that the LET Actor puts a virtual grid on the volume it is attached to. Any changes on the LET Actor will not influence the geometry/material or physics of the particle tranpsort simulation."""

    # hints for IDE
    averaging_method: str
    score_in: str

    user_info_defaults = {
        "averaging_method": (
            "dose_average",
            {
                "doc": "The LET actor returns either dose or fluence (also called track) average. Select the type with this command.",
                "allowed_values": ("dose_average", "track_average"),
            },
        ),
        "score_in": (
            "G4_WATER",
            {
                "doc": "The score_in command allows to convert the LET from the material, "
                "which is defined in the geometry, to any user defined material. "
                "Note that this does not change the material definition in the geometry. "
                "The default value is 'material', which means that no conversion is "
                "performed and the LET to the local material is scored. "
                "You can use any material defined in the simulation "
                "or pre-defined by Geant4 such as 'G4_WATER', "
                "which may be one of the most use cases of this functionality.",
                "setter_hook": _setter_hook_score_in_let_actor,
            },
        ),
        "separate_output": (
            False,
            {
                "doc": "FIXME",
                "deprecated": "Denominator and numerator images are automatically handled and stored. ",
            },
        ),
    }

    user_output_config = {
        "let": {
            "actor_output_class": ActorOutputQuotientMeanImage,
            "interfaces": {
                "numerator": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 0,
                    "active": True,
                    "write_to_disk": False,
                },
                "denominator": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 1,
                    "active": True,
                    "write_to_disk": False,
                },
                "let": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": "quotient",
                    "write_to_disk": True,
                    "active": True,
                    "suffix": None,  # default suffix would be 'let', but we prefer no suffix
                },
            },
        },
    }

    def __init__(self, *args, **kwargs):
        VoxelDepositActor.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateLETActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "BeginOfRunActionMasterThread",
                "EndOfRunActionMasterThread",
                "BeginOfEventAction",
            }
        )

    def initialize(self):
        """
        At the start of the run, the image is centered according to the coordinate system of
        the attached volume. This function computes the correct origin = center + translation.
        Note that there is a half-pixel shift to align according to the center of the pixel,
        like in ITK.
        """
        VoxelDepositActor.initialize(self)

        self.check_user_input()

        self.InitializeUserInfo(self.user_info)
        # Set the physical volume name on the C++ side
        self.SetPhysicalVolumeName(self.get_physical_volume_name())
        self.InitializeCpp()

    def BeginOfRunActionMasterThread(self, run_index):
        self.prepare_output_for_run("let", run_index)
        # self.prepare_output_for_run("let_numerator", run_index)
        # self.prepare_output_for_run("let_denominator", run_index)

        self.push_to_cpp_image(
            "let", run_index, self.cpp_numerator_image, self.cpp_denominator_image
        )
        g4.GateLETActor.BeginOfRunActionMasterThread(self, run_index)

    def EndOfRunActionMasterThread(self, run_index):
        self.fetch_from_cpp_image(
            "let", run_index, self.cpp_numerator_image, self.cpp_denominator_image
        )
        self._update_output_coordinate_system("let", run_index)
        self.user_output.let.store_meta_data(
            run_index, number_of_samples=self.NbOfEvent
        )

        VoxelDepositActor.EndOfRunActionMasterThread(self, run_index)
        return 0

    def EndSimulationAction(self):
        g4.GateLETActor.EndSimulationAction(self)
        VoxelDepositActor.EndSimulationAction(self)


def _setter_hook_lookup_table_path(self, path):
    if not os.path.exists(path):
        raise ValueError(f"Lookup table path {path} does not exist.")
    return path


class BeamQualityActor(VoxelDepositActor, g4.GateBeamQualityActor):
    """
    BeamQualityActor:
    Handles the logic for RE, RBE and potentially other actors

    """

    user_info_defaults = {
        "energy_per_nucleon": (
            False,
            {
                "doc": "The kinetic energy in the table is the energy per nucleon MeV/n, else False.",
            },
        ),
        "model": (
            "RE",
            {
                "doc": "which model is used to calculate beam quality.",
                "allowed_values": ("mMKM", "LEM1lda", "RE"),
            },
        ),
        "score_in": (
            "material",
            {
                "doc": """The score_in command allows to convert the scored quantity from the material, which is defined in the geometry, to any user defined material. Note, that this does not change the material definition in the geometry. The default value is 'material', which means that no conversion is performed and the quantity to the local material is scored. You can use any material defined in the simulation or pre-defined by Geant4 such as 'G4_WATER', which may be one of the most use cases of this functionality.
                """,
            },
        ),
        "lookup_table_path": (
            "",
            {
                "doc": "path of the z*_1d or alpha_z table.",
                "setter_hook": _setter_hook_lookup_table_path,
            },
        ),
        "lookup_table": (
            None,
            {
                "doc": "z*_1d or alpha_z table. Read by the actor.",
            },
        ),
    }
    scored_quantity = "alpha"
    user_output_config = {
        f"{scored_quantity}_mix": {
            "actor_output_class": ActorOutputQuotientMeanImage,
            "interfaces": {
                f"{scored_quantity}_numerator": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 0,
                    "active": True,
                    "write_to_disk": False,
                },
                f"{scored_quantity}_denominator": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 1,
                    "active": True,
                    "write_to_disk": False,
                },
                f"{scored_quantity}_mix": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": "quotient",
                    "write_to_disk": True,
                    "active": True,
                    # "suffix": None,  # default suffix would be 'alpha_mix', but we prefer no suffix
                },
            },
        },
        "beta_mix": {
            "actor_output_class": ActorOutputQuotientMeanImage,
            "interfaces": {
                "beta_numerator": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 0,
                    "active": False,
                    "write_to_disk": False,
                },
                "beta_denominator": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 1,
                    "active": False,
                    "write_to_disk": False,
                },
                "beta_mix": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": "quotient",
                    "write_to_disk": True,
                    "active": False,
                    # "suffix": None,  # default suffix would be 'beta_mix', but we prefer no suffix
                },
            },
        },
    }

    def __init__(self, *args, **kwargs):
        # Init parent
        VoxelDepositActor.__init__(self, *args, **kwargs)
        # calculate some internal variables
        self.element_to_atomic_number_dict = {
            "H": 1,
            "He": 2,
            "Li": 3,
            "Be": 4,
            "B": 5,
            "C": 6,
            "N": 7,
            "O": 8,
            "F": 9,
            "Ne": 10,
        }
        self.atomic_number_to_mass_dict = {
            1: 1,
            2: 4,
            3: 6.9,
            4: 9,
            5: 10.8,
            6: 12,
            7: 14,
            8: 16,
            9: 19,
            10: 20.2,
        }
        self.lookup_table = None
        self._extend_table_to_zero_and_inft = True
        self.multiple_scoring = False
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateBeamQualityActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "BeginOfRunActionMasterThread",
                "EndOfRunActionMasterThread",
                "BeginOfEventAction",
            }
        )

    def initialize(self):
        """
        At the start of the run, the image is centered according to the coordinate system of
        the attached volume. This function computes the correct origin = center + translation.
        Note that there is a half-pixel shift to align according to the center of the pixel,
        like in ITK.
        """

        VoxelDepositActor.initialize(self)

        self.check_user_input()

        if self.lookup_table_path:
            if not Path(self.lookup_table_path).is_file():
                raise ValueError(
                    f"Lookuptable path does not exist or is no file: {self.lookup_table_path}"
                )

            self.read_lookup_table(self.lookup_table_path)

        if self.lookup_table is None:
            raise ValueError(
                "Missing lookup table. Set it manually with the lookup_table attribute or provide a table path to the lookup_table_path attribute."
            )

        if self.multiple_scoring:
            self.user_output.beta_mix.set_active(True, item="all")
            self.user_output.beta_mix.set_write_to_disk(True, item="quotient")

        self.InitializeUserInfo(self.user_info)
        # Set the physical volume name on the C++ side
        self.SetPhysicalVolumeName(self.get_physical_volume_name())
        self.InitializeCpp()

    def read_lookup_table_txt(self, table_path):
        # Element-Z mapping

        with open(table_path, "r") as f:
            lines = f.readlines()
        # add extra line to sign end of file
        lines.append("\n")
        start_table = False
        end_table = True
        fragments = []
        v_table = []
        for line in lines:
            if "Fragment" in line:
                element = line.split()[1]
                if element in self.element_to_atomic_number_dict:
                    Z = self.element_to_atomic_number_dict[element]
                else:
                    raise ValueError(
                        f"Lookup table inconsistency: cannot convert element name to atomic number: {element}"
                    )
                values = []
                energy = []
                v_table.append([Z])
                fragments.append(Z)
                start_table = True
                end_table = False
            elif line.startswith("\n") == False and start_table:
                # if count == 1:  # energy vector is the same for all Z
                energy.append(float(line.split()[0]))
                values.append(float(line.split()[1]))
            elif not end_table:
                start_table = False
                # if count == 1:
                # e_table.append(energy)
                v_table.append(energy)
                v_table.append(values)  # we want to do this only once per table
                end_table = True
        return v_table, fragments

    def read_lookup_table_csv(self, table_path):
        dataset = pd.read_csv(table_path)
        particles = dataset["particle"].unique()
        fragments = []
        v_table = []
        for p in particles:
            if p in self.element_to_atomic_number_dict:
                Z = self.element_to_atomic_number_dict[p]
            else:
                raise ValueError(
                    f"Lookup table inconsistency: cannot convert element name to atomic number: {p}"
                )
            fragments.append(Z)
            v_table.append([Z])
            energies_p = dataset.loc[dataset["particle"] == p, "meanEnergy"]
            v_table.append(list(energies_p))
            values_p = dataset.loc[dataset["particle"] == p, "z1D*"]
            v_table.append(list(values_p))
        return v_table, fragments

    def read_lookup_table(self, table_path):
        p = Path(table_path)
        if p.suffix == ".txt":
            v_table, fragments = self.read_lookup_table_txt(table_path)
        elif p.suffix == ".csv":
            v_table, fragments = self.read_lookup_table_csv(table_path)
        else:
            raise NotImplementedError(
                "Cannot read lookup table. File type reading not implemented yet."
            )

        self.set_lookup_table(v_table, fragments)

    def set_lookup_table(self, v_table: list, fragments: list):
        element_map = self.element_to_atomic_number_dict
        # convert fragments strings to atomic number if they are
        fragments = [
            element_map[element] if element in element_map else element
            for element in fragments
        ]
        # Check if all values in `a` are positive
        if all(isinstance(x, (int, float)) and x > 0 for x in fragments):
            # Convert all elements in `a` to integers
            fragments = [
                int(x) if isinstance(x, (int, float)) else x for x in fragments
            ]
        else:
            raise ValueError("Non numeric entries in table")
        self.check_table(v_table, fragments)
        # get max and min atomic numbers available
        self.ZMinTable = min(fragments)
        self.ZMaxTable = max(fragments)
        if len(v_table) != 3 * len(fragments):
            raise ValueError(
                f"Error: {len(v_table) = } and {len(fragments) = } are not equal."
            )
        if self.energy_per_nucleon:
            for i in range(1, len(v_table), 3):
                n_nuclei = self.atomic_number_to_mass_dict[
                    v_table[i - 1][0]
                ]  # v_table[i - 1][0] * (1 + float(v_table[i - 1][0] != 1))
                v_table[i] = [x * n_nuclei for x in v_table[i]]
        if self._extend_table_to_zero_and_inft:
            for i in range(1, len(v_table), 3):
                energy_0 = v_table[i][0]
                energy_max = v_table[i][-1]
                if energy_0 > 0:
                    v_table[i].insert(0, 0.0)
                    v_table[i + 1].insert(0, v_table[i + 1][0])
                if energy_max < np.finfo(np.float32).max:
                    # we use a single precision float here to enable on the cpp side the table to be a float instead of a double for more efficient cache use --> future improvmement
                    v_table[i].append(np.finfo(np.float32).max)
                    v_table[i + 1].append(v_table[i + 1][-1])

        self.lookup_table = v_table

    def check_table(self, v_table, fragments):
        if not fragments:
            raise ValueError("Fragment table is empty")
        sorted_fragments = sorted(fragments)
        if not sorted_fragments == list(
            range(sorted_fragments[0], sorted_fragments[0] + len(sorted_fragments))
        ):
            raise ValueError("Fragment list is missing entries")

    def BeginOfRunActionMasterThread(self, run_index):

        self.prepare_output_for_run(f"{self.scored_quantity}_mix", run_index)
        self.push_to_cpp_image(
            f"{self.scored_quantity}_mix",
            run_index,
            self.cpp_numerator_alpha_image,
            self.cpp_denominator_image,
        )

        if self.multiple_scoring:
            self.prepare_output_for_run("beta_mix", run_index)
            self.push_to_cpp_image(
                "beta_mix",
                run_index,
                self.cpp_numerator_beta_image,
                self.cpp_denominator_image,
            )

        g4.GateBeamQualityActor.BeginOfRunActionMasterThread(self, run_index)

    def EndOfRunActionMasterThread(self, run_index):
        self.fetch_from_cpp_image(
            f"{self.scored_quantity}_mix",
            run_index,
            self.cpp_numerator_image,
            self.cpp_denominator_image,
        )
        self._update_output_coordinate_system(f"{self.scored_quantity}_mix", run_index)
        self.user_output.__getattr__(f"{self.scored_quantity}_mix").store_meta_data(
            run_index, number_of_samples=self.NbOfEvent
        )
        if self.multiple_scoring:
            self.fetch_from_cpp_image(
                "beta_mix",
                run_index,
                self.cpp_numerator_beta_image,
                self.cpp_denominator_image,
            )
            self._update_output_coordinate_system("beta_mix", run_index)
            self.user_output.beta_mix.store_meta_data(
                run_index, number_of_samples=self.NbOfEvent
            )

        VoxelDepositActor.EndOfRunActionMasterThread(self, run_index)
        return 0

    def EndSimulationAction(self):
        g4.GateBeamQualityActor.EndSimulationAction(self)
        VoxelDepositActor.EndSimulationAction(self)

    def compute_dose_from_edep_img(self, overrides=dict()):
        """
        * cretae mass image:
            - from ct HU units, if dose actor attached to ImageVolume.
            - from material density, if standard volume
        * compute dose as edep_image /  mass_image
        """
        vol_name = self.user_info.attached_to
        vol = self.simulation.volume_manager.get_volume(vol_name)
        vol_type = vol.volume_type

        spacing = np.array(self.user_info.spacing)
        voxel_volume = spacing[0] * spacing[1] * spacing[2]
        Gy = g4_units.Gy

        edep_img = (
            self.user_output.__getattr__(f"{self.scored_quantity}_mix")
            .merged_data.data[1]
            .image
        )

        score_in_material = self.user_info.score_in

        material_database = (
            self.simulation.volume_manager.material_database.g4_materials
        )

        if score_in_material != "material":
            # score in other material
            if score_in_material not in material_database:
                self.simulation.volume_manager.material_database.FindOrBuildMaterial(
                    score_in_material
                )
            density = material_database[score_in_material].GetDensity()
            dose_img = scale_itk_image(edep_img, 1 / (voxel_volume * density * Gy))
        else:
            # divide edep image with the original mass image
            if vol_type == "ImageVolume":
                density_img = vol.create_density_image()
                edep_density = divide_itk_images(
                    img1_numerator=edep_img,
                    img2_denominator=density_img,
                    filterVal=0,
                    replaceFilteredVal=0,
                )
                # divide by voxel volume and convert unit.
                dose_img = scale_itk_image(
                    edep_density, g4_units.cm3 / (Gy * voxel_volume * g4_units.g)
                )
            else:
                if vol.material not in material_database:
                    self.simulation.volume_manager.material_database.FindOrBuildMaterial(
                        vol.material
                    )
                density = material_database[vol.material].GetDensity()
                dose_img = scale_itk_image(edep_img, 1 / (voxel_volume * density * Gy))

        dose_image = ItkImageDataItem(data=dose_img)
        dose_image.copy_image_properties(edep_img)

        return dose_image


class REActor(BeamQualityActor, g4.GateBeamQualityActor):
    scored_quantity = "RE"
    user_output_config = {
        f"{scored_quantity}_mix": {
            "actor_output_class": ActorOutputQuotientMeanImage,
            "interfaces": {
                f"{scored_quantity}_numerator": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 0,
                    "active": True,
                    "write_to_disk": False,
                },
                f"{scored_quantity}_denominator": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": 1,
                    "active": True,
                    "write_to_disk": False,
                },
                f"{scored_quantity}_mix": {
                    "interface_class": UserInterfaceToActorOutputImage,
                    "item": "quotient",
                    "write_to_disk": True,
                    "active": True,
                    # "suffix": None,  # default suffix would be 'alpha_mix', but we prefer no suffix
                },
            },
        },
    }


class RBEActor(BeamQualityActor, g4.GateBeamQualityActor):
    """
    RBEActor:
        - mMKM:
            - alpha mix (score separately numerator and denominator)
        - LEM1lda:
            - alpha mix (score separately numerator and denominator)
            - beta mix (score separately numerator and denominator)

        Note: all will also score edep as denominator image -> retrieve dose
    """

    user_info_defaults = {
        "alpha_0": (
            0.172,
            {
                "doc": "mMKM specific fitted parameter, for the calculation of cellline alpha_z per radiation. Unit: Gy-1",
            },
        ),
        "beta_ref": (
            None,
            {
                "doc": "Cellline specific parameter. Initialized according to cell_type.",
            },
        ),
        "F_clin": (
            2.41,
            {
                "doc": "mMKM specific arbituary parameter, the clinical sclaing factor.",
            },
        ),
        "D_cut": (
            30.0,
            {
                "doc": "LEM specific arbituary parameter, the physical dose threshold between linear quadratic and linear cell survival. Unit: Gy.",
            },
        ),
        "r_nucleus": (
            5,
            {
                "doc": "nucleus's radius, in um. Used when rbe_model is LEM",
            },
        ),
        "cell_type": (
            "HSG",
            {
                "doc": "To add new cell types, update the self.cells_radiosensitivity"
                "variable in the init of the actor",
                "allowed_values": ("HSG", "Chordoma"),
            },
        ),
        "write_RBE_dose_image": (
            True,
            {
                "doc": "Do you want to calcu;ate and write to disk RBE and RBE dose images?",
            },
        ),
    }

    user_output_config = BeamQualityActor.user_output_config.copy()
    user_output_config.update(
        {
            "rbe": {
                "actor_output_class": ActorOutputSingleMeanImage,
                "active": False,
            },
            "rbe_dose": {
                "actor_output_class": ActorOutputSingleMeanImage,
                "active": False,
            },
        }
    )

    def __init__(self, *args, **kwargs):
        # Init parent
        BeamQualityActor.__init__(self, *args, **kwargs)

        # internal variables
        self.cells_radiosensitivity = {
            "HSG": {"alpha_ref": 0.764, "beta_ref": 0.0615},
            "Chordoma": {"alpha_ref": 0.1, "beta_ref": 0.05},
        }

        self.s_max = None
        self.lnS_cut = None

        self.__initcpp__()

    def initialize(self):
        """
        At the start of the run, the image is centered according to the coordinate system of
        the attached volume. This function computes the correct origin = center + translation.
        Note that there is a half-pixel shift to align according to the center of the pixel,
        like in ITK.
        """
        VoxelDepositActor.initialize(self)

        self.check_user_input()

        if self.lookup_table_path:
            self.read_lookup_table(self.lookup_table_path)

        if self.lookup_table is None:
            raise ValueError(
                "Missing lookup table. Set it manually with the lookup_table attribute or provide a table path to the lookup_table_path attribute."
            )

        # calculate some internal variables
        self.fAreaNucl = (np.pi * self.r_nucleus**2) * g4_units.um * g4_units.um
        alpha_ref = self.cells_radiosensitivity[self.cell_type]["alpha_ref"]
        beta_ref = self.cells_radiosensitivity[self.cell_type]["beta_ref"]
        self.beta_ref = beta_ref
        self.s_max = alpha_ref + 2 * beta_ref * self.D_cut
        self.fSmax = self.s_max  # set also to cpp
        self.lnS_cut = -beta_ref * self.D_cut**2 - alpha_ref * self.D_cut

        if self.model == "LEM1lda":
            self.multiple_scoring = True
            self.user_output.beta_mix.set_active(True, item="all")
            self.user_output.beta_mix.set_write_to_disk(True, item="quotient")

        self.InitializeUserInfo(self.user_info)
        # Set the physical volume name on the C++ side
        self.SetPhysicalVolumeName(self.get_physical_volume_name())
        self.InitializeCpp()

    def EndSimulationAction(self):
        g4.GateBeamQualityActor.EndSimulationAction(self)
        if self.model == "mMKM":
            self._postprocess_alpha_numerator_mkm()
        if self.write_RBE_dose_image:
            self.compute_rbe_weighted_dose()
        VoxelDepositActor.EndSimulationAction(self)

    def _postprocess_alpha_numerator_mkm(self):
        beta_ref = self.cells_radiosensitivity[self.cell_type]["beta_ref"]
        alpha_mix_numerator_img = self.user_output.__getattr__(
            f"{self.scored_quantity}_mix"
        ).merged_data.data[0]
        alpha_mix_denominator_img = (
            self.user_output.__getattr__(
                f"{self.scored_quantity}_mix"
            ).merged_data.data[1]
            * self.alpha_0
        )
        self.user_output.__getattr__(f"{self.scored_quantity}_mix").merged_data.data[
            0
        ] = (alpha_mix_numerator_img * beta_ref + alpha_mix_denominator_img)

    def compute_rbe_weighted_dose(self):
        alpha_ref = self.cells_radiosensitivity[self.cell_type]["alpha_ref"]
        beta_ref = self.cells_radiosensitivity[self.cell_type]["beta_ref"]
        dose_img = self.compute_dose_from_edep_img()

        if self.model == "mMKM":
            alpha_mix_img = self.user_output.__getattr__(
                f"{self.scored_quantity}_mix"
            ).merged_data.quotient
            log_survival = alpha_mix_img * dose_img * (
                -1
            ) + dose_img * dose_img * beta_ref * (-1)
            log_survival_arr = log_survival.image_array
        elif self.model == "LEM1lda":
            dose_arr = dose_img.image_array
            arr_mask_linear = dose_arr > self.D_cut
            alpha_mix_img = self.user_output.__getattr__(
                f"{self.scored_quantity}_mix"
            ).merged_data.quotient
            sqrt_beta_mix_img = self.user_output.beta_mix.merged_data.quotient

            log_survival_lq = alpha_mix_img * dose_img * (
                -1
            ) + dose_img * dose_img * sqrt_beta_mix_img * sqrt_beta_mix_img * (-1)
            log_survival_lq_arr = log_survival_lq.image_array
            log_survival_linear = (
                alpha_mix_img * self.D_cut * (-1)
                + sqrt_beta_mix_img * sqrt_beta_mix_img * self.D_cut * self.D_cut * (-1)
                + (dose_img + self.D_cut * (-1)) * self.s_max * (-1)
            )
            log_survival_linear_arr = log_survival_linear.image_array

            log_survival_arr = np.zeros(log_survival_lq.image_array.shape)
            log_survival_arr[arr_mask_linear] = log_survival_linear_arr[arr_mask_linear]
            log_survival_arr[~arr_mask_linear] = log_survival_lq_arr[~arr_mask_linear]

        # solve linear quadratic equation to get Dx
        if self.model == "mMKM":
            rbe_dose_arr = (
                (-alpha_ref + np.sqrt(alpha_ref**2 - 4 * beta_ref * log_survival_arr))
                / (2 * beta_ref)
                * self.F_clin
            )
        else:
            arr_mask_linear = log_survival_arr < self.lnS_cut
            rbe_dose_lq_arr = (
                -alpha_ref + np.sqrt(alpha_ref**2 - 4 * beta_ref * log_survival_arr)
            ) / (2 * beta_ref)
            rbe_dose_linear_arr = (
                -log_survival_arr + self.lnS_cut
            ) / self.s_max + self.D_cut
            rbe_dose_arr = np.zeros(log_survival_arr.shape)
            rbe_dose_arr[arr_mask_linear] = rbe_dose_linear_arr[arr_mask_linear]
            rbe_dose_arr[~arr_mask_linear] = rbe_dose_lq_arr[~arr_mask_linear]

        # create new data item for the survival image. Same metadata as the other images, but new image array
        rbedose_img = itk_image_from_array(rbe_dose_arr)
        rbe_weighted_dose_image = ItkImageDataItem(data=rbedose_img)
        rbe_weighted_dose_image.copy_image_properties(dose_img.image)

        rbe_image = rbe_weighted_dose_image / dose_img

        self.user_output.rbe.merged_data = rbe_image
        rbe_path = self.user_output.rbe.get_output_path()

        self.user_output.rbe_dose.merged_data = rbe_weighted_dose_image
        rbe_dose_path = self.user_output.rbe_dose.get_output_path()

        rbe_weighted_dose_image.write(rbe_dose_path)
        rbe_image.write(rbe_path)


class ProductionAndStoppingActor(VoxelDepositActor, g4.GateProductionAndStoppingActor):
    """This actor scores the number of particles stopping or starting in a voxel grid in the volume to which the actor is attached to."""

    # hints for IDE
    method: str

    user_info_defaults = {
        "method": (
            "stopping",
            {
                "doc": "Want to score production or stopping of particles?",
                "allowed_values": ("production", "stopping"),
            },
        )
    }

    user_output_config = {
        "production_stopping": {
            "actor_output_class": ActorOutputSingleImage,
        },
    }

    def __init__(self, *args, **kwargs):
        VoxelDepositActor.__init__(self, *args, **kwargs)

        self.__initcpp__()

    def __initcpp__(self):
        g4.GateProductionAndStoppingActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "BeginOfRunActionMasterThread",
                "EndOfRunActionMasterThread",
                "BeginOfEventAction",
            }
        )

    def initialize(self):
        """
        At the start of the run, the image is centered according to the coordinate system of
        the attached volume. This function computes the correct origin = center + translation.
        Note that there is a half-pixel shift to align according to the center of the pixel,
        like in ITK.
        """
        VoxelDepositActor.initialize(self)

        self.check_user_input()

        self.InitializeUserInfo(self.user_info)
        # Set the physical volume name on the C++ side
        self.SetPhysicalVolumeName(self.get_physical_volume_name())
        self.InitializeCpp()

    def BeginOfRunActionMasterThread(self, run_index):
        self.prepare_output_for_run("production_stopping", run_index)

        self.push_to_cpp_image("production_stopping", run_index, self.cpp_value_image)
        g4.GateProductionAndStoppingActor.BeginOfRunActionMasterThread(self, run_index)

    def EndOfRunActionMasterThread(self, run_index):
        self.fetch_from_cpp_image(
            "production_stopping", run_index, self.cpp_value_image
        )
        self._update_output_coordinate_system("production_stopping", run_index)
        self.user_output.production_stopping.store_meta_data(run_index)

        VoxelDepositActor.EndOfRunActionMasterThread(self, run_index)
        return 0

    def EndSimulationAction(self):
        g4.GateProductionAndStoppingActor.EndSimulationAction(self)
        VoxelDepositActor.EndSimulationAction(self)


class FluenceActor(VoxelDepositActor, g4.GateFluenceActor):
    """
    FluenceActor: compute a 3D map of fluence
    FIXME: add scatter order and uncertainty
    """

    # hints for IDE
    uncertainty: bool
    scatter: bool

    user_info_defaults = {
        "uncertainty": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "scatter": (
            False,
            {
                "doc": "FIXME",
            },
        ),
    }

    user_output_config = {
        "fluence": {
            "actor_output_class": ActorOutputSingleImage,
        },
    }

    def __init__(self, *args, **kwargs):
        VoxelDepositActor.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateFluenceActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "BeginOfRunActionMasterThread",
                "EndOfRunActionMasterThread",
                "BeginOfEventAction",
            }
        )

    def initialize(self):
        VoxelDepositActor.initialize(self)

        self.check_user_input()

        # no options yet
        if self.uncertainty or self.scatter:
            fatal("FluenceActor : uncertainty and scatter not implemented yet")

        self.InitializeUserInfo(self.user_info)
        # Set the physical volume name on the C++ side
        self.SetPhysicalVolumeName(self.get_physical_volume_name())
        self.InitializeCpp()

    def BeginOfRunActionMasterThread(self, run_index):
        self.prepare_output_for_run("fluence", run_index)
        self.push_to_cpp_image("fluence", run_index, self.cpp_fluence_image)
        g4.GateFluenceActor.BeginOfRunActionMasterThread(self, run_index)

    def EndOfRunActionMasterThread(self, run_index):
        self.fetch_from_cpp_image("fluence", run_index, self.cpp_fluence_image)
        self._update_output_coordinate_system("fluence", run_index)
        self.user_output.fluence.store_meta_data(
            run_index, number_of_samples=self.NbOfEvent
        )
        VoxelDepositActor.EndOfRunActionMasterThread(self, run_index)
        return 0

    def EndSimulationAction(self):
        g4.GateFluenceActor.EndSimulationAction(self)
        VoxelDepositActor.EndSimulationAction(self)


class EmCalculatorActor(ActorBase, g4.GateEmCalculatorActor):
    user_info_defaults = {
        "is_ion": (
            True,
            {
                "doc": "Wheather the particle we calculate dedx for is an ion",
            },
        ),
        "particle_name": (
            "",
            {
                "doc": "GenericIon if the particle is ion, else name of the particle",
            },
        ),
        "ion_params": (
            "",
            {
                "doc": "string indicating atomic number and mass, separated by one space. Example '6 12' for C12",
            },
        ),
        "material": (
            "",
            {
                "doc": "Material in which to score dedx",
            },
        ),
        "nominal_energies": (
            [],
            {
                "doc": "List of nominal energies to use to generate the dedx table",
            },
        ),
        "savefile_path": (
            "",
            {
                "doc": "file path to save the table",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        ActorBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateEmCalculatorActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "BeginOfRunActionMasterThread",
                "EndOfRunActionMasterThread",
                "BeginOfRunAction",
                "EndOfRunAction",
                "BeginOfEventAction",
                "SteppingAction",
            }
        )

    def initialize(self, *args):
        if self.is_ion:
            self.particle_name = "GenericIon"
        self.InitializeUserInfo(self.user_info)  # C++ side
        self.InitializeCpp()


process_cls(VoxelDepositActor)
process_cls(DoseActor)
process_cls(TLEDoseActor)
process_cls(LETActor)
process_cls(RBEActor)
process_cls(REActor)
process_cls(BeamQualityActor)
process_cls(FluenceActor)
process_cls(ProductionAndStoppingActor)
process_cls(EmCalculatorActor)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test028_ge_nm670_spect_2_helpers as test028
from opengate.tests import utility
from pathlib import Path

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "", output_folder="test028_proj_mt"
    )

    # create the simulation
    sim = gate.Simulation()
    cm = gate.g4_units.cm

    # main description
    spect = test028.create_spect_simu(sim, paths, 4)
    proj = test028.test_add_proj(sim)

    # rotate spect
    test028.nm670.rotate_gantry(spect, 10 * cm, -15)

    # go
    sim.run()

    # check
    proj_out = sim.get_actor("Projection")
    output_ref_folder = Path(str(paths.output_ref).replace("_mt", ""))
    output_ref_filename = "proj028_counts.mhd".replace("_mt", "")
    print(output_ref_folder)
    print(output_ref_filename)
    is_ok = test028.test_spect_proj(
        sim, paths, proj_out, output_ref_folder, output_ref_filename
    )

    utility.test_ok(is_ok)



import opengate as gate

from scipy.spatial.transform import Rotation
import gatetools.phsp as phsp
import os
from opengate.tests import utility
from opengate.contrib.tps.treatmentPlanPhsSource import TreatmentPlanPhsSource
from opengate.contrib.tps.ionbeamtherapy import spots_info_from_txt
from pathlib import Path

# units
m = gate.g4_units.m
mm = gate.g4_units.mm
cm = gate.g4_units.cm
nm = gate.g4_units.nm
Bq = gate.g4_units.Bq
MeV = gate.g4_units.MeV
deg: float = gate.g4_units.deg


def create_test_phsp(
    particle="proton",
    phs_name=Path("output") / "test_proton.root",
    number_of_particles=1,
    translation=None,
):
    # create the simulation
    if translation is None:
        translation = [0 * mm, 0 * mm, 0 * mm]
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    # ui.visu = True
    ui.visu_type = "vrml"
    ui.check_volumes_overlap = False
    # ui.running_verbose_level = gate.EVENT
    ui.number_of_threads = 1
    ui.random_seed = "auto"

    ##########################################################################################
    # geometry
    ##########################################################################################
    #  adapt world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_Galactic"

    # virtual plane for phase space
    plane = sim.add_volume("Tubs", "phase_space_plane")
    # plane.material = "G4_AIR"
    plane.material = "G4_Galactic"
    plane.rmin = 0
    plane.rmax = 30 * cm
    plane.dz = 1 * nm  # half height
    # plane.rotation = Rotation.from_euler("xy", [180, 30], degrees=True).as_matrix()
    # plane.translation = [0 * mm, 0 * mm, 0 * mm]
    plane.translation = translation

    plane.color = [1, 0, 0, 1]  # red

    ##########################################################################################
    # Actors
    ##########################################################################################
    # Split the joined path into the directory path and filename
    directory_path, filename = os.path.split(phs_name)
    # Extract the base filename and extension
    base_filename, extension = os.path.splitext(filename)
    new_extension = ".root"

    # PhaseSpace Actor
    ta1 = sim.add_actor("PhaseSpaceActor", "PhaseSpace1")
    ta1.attached_to = plane
    ta1.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "ParticleName",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
    ]
    new_joined_path = os.path.join(directory_path, base_filename + new_extension)
    ta1.output = new_joined_path
    ta1.debug = False
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = particle
    ta1.filters.append(f)

    sim.physics_manager.physics_list_name = "QGSP_BIC_EMZ"

    ##########################################################################################
    #  Source
    ##########################################################################################
    source = sim.add_source("GenericSource", "particle_source")
    source.attached_to = "world"
    # source.particle = "ion 6 12"  # Carbon ions
    source.particle = "proton"
    source.position.type = "point"
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.position.translation = [0 * cm, 0 * cm, -0.1 * cm]
    source.energy.type = "mono"
    source.energy.mono = 150 * MeV
    source.n = number_of_particles

    sim.run()
    output = sim.output


def create_phsp_without_source(
    output_dir=Path("output"),
    phs_name="test_proton.root",
):
    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # ui.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    # ui.running_verbose_level = gate.EVENT
    sim.number_of_threads = 1
    sim.random_seed = "auto"
    sim.output_dir = output_dir

    ##########################################################################################
    # geometry
    ##########################################################################################
    #  adapt world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_Galactic"
    print(world.name)

    # virtual plane for phase space
    plane = sim.add_volume("Tubs", "phase_space_plane")
    # plane.material = "G4_AIR"
    plane.material = "G4_Galactic"
    plane.rmin = 0
    plane.rmax = 30 * cm
    plane.dz = 1 * nm  # half height
    # plane.rotation = Rotation.from_euler("xy", [180, 30], degrees=True).as_matrix()
    plane.translation = [0 * mm, 0 * mm, 0 * mm]
    # plane.translation = translation

    plane.color = [1, 0, 0, 1]  # red

    ##########################################################################################
    # Actors
    ##########################################################################################

    # PhaseSpace Actor
    ta1 = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    ta1.attached_to = plane
    ta1.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "ParticleName",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
    ]
    ta1.output_filename = phs_name
    ta1.debug = False

    sim.physics_manager.physics_list_name = "QGSP_BIC_EMZ"

    # ##########################################################################################
    # #  Source
    # ##########################################################################################
    # # phsp source
    # source = sim.add_source("PhaseSpaceSource", "phsp_source_global")
    # source.attached_to = world.name
    # source.phsp_file = source_name
    # source.position_key = "PrePosition"
    # source.direction_key = "PreDirection"
    # source.global_flag = True
    # source.particle = particle
    # source.batch_size = 3000
    # source.n = number_of_particles / ui.number_of_threads
    # # source.position.translation = [0 * cm, 0 * cm, -35 * cm]

    # output = sim.run()
    # output = sim.start()
    # output = sim.start(start_new_process=True)
    return sim, plane


def test_source_rotation_a(
    plan_file_name=Path("output") / "test_proton_offset.root",
    phs_list_file_name="PhsList.txt",
    phs_folder_name="",
    output_dir=Path("output"),
    phs_file_name_out="test_source_electron.root",
) -> gate.Simulation:
    sim, plane = create_phsp_without_source(
        phs_name=phs_file_name_out, output_dir=output_dir
    )
    number_of_particles = 1
    ##########################################################################################
    #  Source
    ##########################################################################################
    # TreatmentPlanPhsSource source
    tpPhSs = TreatmentPlanPhsSource("RT_plan", sim)
    tpPhSs.set_phaseSpaceList_file_name(phs_list_file_name)
    tpPhSs.set_phaseSpaceFolder(phs_folder_name)
    beam_dict = spots_info_from_txt(plan_file_name, "", 1)
    G = beam_dict["gantry_angle"]
    spots = beam_dict["spots"]
    tpPhSs.set_spots(spots)
    tpPhSs.set_particles_to_simulate(number_of_particles)
    tpPhSs.set_distance_source_to_isocenter(100 * cm)
    tpPhSs.set_distance_stearmag_to_isocenter(5 * m, 5 * m)
    tpPhSs.rotation = Rotation.from_euler("z", G, degrees=True)
    tpPhSs.initialize_tpPhssource()

    # depending on the rotation of the gantry, the rotation of the phase space to catch the particles is different
    plane.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()

    sim.run()

    return sim


def get_first_entry_of_key(
    file_name_root=Path("output") / "test_source_electron.root", key="ParticleName"
):
    # read root file
    data_ref, keys_ref, m_ref = phsp.load(file_name_root)
    index = keys_ref.index(key)
    return data_ref[0][index]


def check_value_from_root_file(
    file_name_root=Path("output") / "test_source_electron.root",
    key="ParticleName",
    ref_value="e-",
):
    # read root file
    value = get_first_entry_of_key(file_name_root=file_name_root, key=key)
    if (type(ref_value) != str) and (type(value) != str):
        is_ok = utility.check_diff_abs(
            float(value), float(ref_value), tolerance=1e-3, txt=key
        )
    else:
        if value == ref_value:
            is_ok = True
        else:
            is_ok = False
    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from opengate.tests.src.test081_tle_helpers import add_source
import numpy as np

from opengate.tests.utility import print_test

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test081_tle")

    # create the simulation
    sim = gate.Simulation()

    # main options
    # sim.visu = True
    sim.visu_type = "qt"
    sim.random_seed = 123654
    sim.output_dir = paths.output
    sim.progress_bar = True
    sim.number_of_threads = 1

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # insert voxelized waterbox
    spacing = 8
    waterbox = sim.add_volume("Image", "waterbox")
    fn = paths.data / "test081_tle" / f"waterbox_with_inserts_{spacing}mm_"
    waterbox.image = f"{fn}image.mhd"
    waterbox.set_materials_from_voxelisation(f"{fn}labels.json")
    waterbox_size = [30 * cm, 30 * cm, 20 * cm]

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.global_production_cuts.all = 1 * mm

    # default source for tests
    source = add_source(sim, n=1e5)

    # add conventional dose actor
    dose_actor = sim.add_actor("DoseActor", "dose_actor")
    dose_actor.output_filename = "test081_vox_speed.mhd"
    dose_actor.attached_to = waterbox
    dose_actor.dose_uncertainty.active = True
    dose_actor.dose.active = True
    dose_actor.size = [100, 100, 100]
    dose_actor.spacing = [x / y for x, y in zip(waterbox_size, dose_actor.size)]

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # start simulation
    sim.run(start_new_process=True)
    pps1 = stats.user_output.stats.pps

    # add tle dose actor
    tle_dose_actor = sim.add_actor("TLEDoseActor", "tle_dose_actor")
    tle_dose_actor.output_filename = "test081_vox_speed_tle.mhd"
    tle_dose_actor.attached_to = waterbox
    tle_dose_actor.dose_uncertainty.active = True
    tle_dose_actor.dose.active = True
    tle_dose_actor.size = dose_actor.size
    tle_dose_actor.spacing = dose_actor.spacing

    # remove first actor
    sim.actor_manager.remove_actor("dose_actor")

    # start simulation
    sim.run()
    pps2 = stats.user_output.stats.pps

    print()
    r = np.fabs(pps1 - pps2) / pps2
    tol = 0.4
    b = r < tol
    print_test(
        b, f"Speed PPS is {pps1} vs {pps2} = {r*100:.2f}% (tol={tol:.2f}) ==> {b}"
    )

    is_ok = b
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect.ge_discovery_nm670 as gate_spect
from opengate.tests import utility
from opengate.sources.utility import get_spectrum

paths = utility.get_default_test_paths(__file__, "gate_test043_garf", "test043")

m = gate.g4_units.m
cm = gate.g4_units.cm
mm = gate.g4_units.mm
nm = gate.g4_units.nm
km = gate.g4_units.km
gcm3 = gate.g4_units.g_cm3
MeV = gate.g4_units.MeV
keV = gate.g4_units.keV
Bq = gate.g4_units.Bq
kBq = 1000 * Bq


def sim_set_world(sim):
    # world size
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    return world


def sim_add_detector_plane(sim, spect_name, distance, plane_name="detPlane"):
    # detector input plane
    detector_plane = sim.add_volume("Box", plane_name)
    detector_plane.mother = spect_name
    detector_plane.size = [57.6 * cm, 44.6 * cm, 1 * nm]
    detector_plane.translation = [0, 0, distance]
    detector_plane.material = "G4_Galactic"
    detector_plane.color = [0, 1, 0, 1]

    return detector_plane


def sim_phys(sim):
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.global_production_cuts.all = 1 * km


def sim_source_test(sim, activity):
    tc99m = get_spectrum("Tc99m", "gamma")
    e = tc99m.energies
    w = tc99m.weights

    # first sphere
    s1 = sim.add_source("GenericSource", "s1")
    s1.particle = "gamma"
    s1.activity = activity * 0.0001
    s1.position.type = "sphere"
    s1.position.radius = 15 * mm
    s1.position.translation = [0, 0, 0]
    s1.direction.type = "momentum"
    s1.direction.momentum = [0, 0, -1]
    s1.energy.type = "spectrum_discrete"
    s1.energy.spectrum_energies = e
    s1.energy.spectrum_weights = w

    # second sphere
    s2 = sim.add_source("GenericSource", "s2")
    s2.particle = "gamma"
    s2.activity = activity * 2
    s2.position.type = "sphere"
    s2.position.radius = 15 * mm
    s2.position.translation = [15 * cm, 0, 0]
    s2.direction.type = "iso"
    s2.energy.type = "spectrum_discrete"
    s2.energy.spectrum_energies = e
    s2.energy.spectrum_weights = w

    # third sphere
    s3 = sim.add_source("GenericSource", "s3")
    s3.particle = "gamma"
    s3.activity = activity
    s3.position.type = "sphere"
    s3.position.radius = 28 * mm
    s3.position.translation = [-10 * cm, 5 * cm, 0]
    s3.direction.type = "iso"
    s3.energy.type = "spectrum_discrete"
    s3.energy.spectrum_energies = e
    s3.energy.spectrum_weights = w

    return s1, s2, s3


def create_sim_test_region(sim):
    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.number_of_threads = 1
    sim.visu = False
    sim.random_seed = 321654987

    # activity
    activity = 1e3 * Bq / sim.number_of_threads

    # add a material database
    sim.volume_manager.add_material_database(paths.gate_data / "GateMaterials.db")

    # init world
    sim_set_world(sim)

    # fake spect head
    head = gate_spect.add_fake_spect_head(sim, "spect")
    head.translation = [0, 0, -15 * cm]

    # detector input plane (+ 1nm to avoid overlap)
    pos, crystal_dist, psd = gate_spect.get_plane_position_and_distance_to_crystal(
        "lehr"
    )
    pos += 1 * nm
    print(f"plane position     {pos / mm} mm")
    print(f"crystal distance   {crystal_dist / mm} mm")
    detPlane = sim_add_detector_plane(sim, head.name, pos)

    sim.physics_manager.set_production_cut("world", "all", 1e3 * m)
    sim.physics_manager.set_production_cut("spect", "all", 1 * mm)

    # physics
    sim_phys(sim)

    # sources
    sim_source_test(sim, activity)

    # arf actor
    arf = sim.add_actor("ARFActor", "arf")
    arf.attached_to = detPlane.name
    arf.output_filename = paths.output / "test043_projection_garf.mhd"
    arf.batch_size = 2e5
    arf.image_size = [128, 128]
    arf.image_spacing = [4.41806 * mm, 4.41806 * mm]
    arf.verbose_batch = True
    arf.distance_to_crystal = crystal_dist  # 74.625 * mm
    arf.distance_to_crystal = 74.625 * mm
    arf.pth_filename = paths.gate_data / "pth" / "arf_Tc99m_v3.pth"
    arf.enable_hit_slice = True
    arf.gpu_mode = (
        utility.get_gpu_mode_for_tests()
    )  # should be "auto" but "cpu" for macOS github actions to avoid mps errors

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "stats")
    s.track_types_flag = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation
from opengate.tests import utility
import itk
import numpy as np


# WARNING: This test does currently not work correctly
# because the tested functionality od the dose actor is currently not available
# Test should be successful, but that is meaningless


def define_run_timing_intervals(n):
    sec = gate.g4_units.second
    start = 0
    end = 1 * sec / n
    run_timing_intervals = []
    for r in range(n):
        run_timing_intervals.append([start, end])
        start = end
        end += 1 * sec / n

    return run_timing_intervals


def calculate_mean(edep_arr, unc_arr, edep_thresh_rel=0.7):
    edep_max = np.amax(edep_arr)
    mask = edep_arr > edep_max * edep_thresh_rel
    unc_used = unc_arr[mask]
    unc_mean = np.mean(unc_used)

    return unc_mean


def run_simulation(n_runs, n_part_tot, n_threads, uncertainty_type="uncertainty"):
    paths = utility.get_default_test_paths(
        __file__, "gate_test029_volume_time_rotation", "test066"
    )

    run_timing_intervals = define_run_timing_intervals(n_runs)

    print(f"--- N runs: {len(run_timing_intervals)* n_threads}")

    n_part_thread = int(n_part_tot / n_threads)

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.random_seed = 983456
    sim.number_of_threads = n_threads
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV

    #  change world size
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # add a simple fake volume to test hierarchy
    # translation and rotation like in the Gate macro
    fake = sim.add_volume("Box", "fake")
    fake.size = [40 * cm, 40 * cm, 40 * cm]
    fake.translation = [1 * cm, 2 * cm, 3 * cm]
    fake.material = "G4_AIR"
    fake.color = [1, 0, 1, 1]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.mother = "fake"
    waterbox.size = [10 * cm, 10 * cm, 10 * cm]
    waterbox.translation = [-3 * cm, -2 * cm, -1 * cm]
    waterbox.rotation = Rotation.from_euler("y", -20, degrees=True).as_matrix()
    waterbox.material = "G4_WATER"
    waterbox.color = [0, 0, 1, 1]

    # physics
    sim.physics_manager.set_production_cut("world", "all", 700 * um)

    # default source for tests
    # the source is fixed at the center, only the volume will move
    source = sim.add_source("GenericSource", "mysource")
    source.energy.mono = 90 * MeV
    source.particle = "proton"
    source.position.type = "disc"
    source.position.radius = 5 * mm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = n_part_thread  # 1 part/s

    # add dose actor
    dose = sim.add_actor("DoseActor", "dose")
    dose.output_filename = "test066-edep.mhd"
    dose.attached_to = waterbox
    dose.size = [40, 40, 40]
    mm = gate.g4_units.mm
    dose.spacing = [2.5 * mm, 2.5 * mm, 2.5 * mm]
    if uncertainty_type == "uncertainty":
        dose.edep_uncertainty.active = True
        # dose.ste_of_mean = False  # currently not available
    elif uncertainty_type == "ste_of_mean":
        dose.edep_uncertainty.active = False
        # dose.ste_of_mean = True  # currently not available

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.track_types_flag = True
    stat.output_filename = "stats066.txt"

    # motion
    sim.run_timing_intervals = run_timing_intervals

    # start simulation
    sim.run(start_new_process=True)

    # print results at the end
    print(stat)

    print(dose)

    # edep_path = dose.edep.get_output_path()
    # unc_path = dose.edep_uncertainty.get_output_path()

    # test that the simulation didn't stop because we reached the planned number of runs
    stats_ref = utility.read_stat_file(paths.output / "stats066.txt")
    n_runs_planned = len(run_timing_intervals) * n_threads
    n_effective_runs = stats_ref.counts.runs
    print(f"{n_runs_planned = }")
    print(f"{n_effective_runs = }")

    return dose.edep.get_data(), dose.edep_uncertainty.get_data()
    # return itk.imread(str(edep_path)), itk.imread(str(unc_path))


if __name__ == "__main__":
    n_part_tot = 10000
    ok = True
    # Uncertainty
    edep_ref, unc_ref = run_simulation(
        n_runs=1, n_part_tot=n_part_tot, n_threads=1, uncertainty_type="uncertainty"
    )
    edep_test, unc_test = run_simulation(
        n_runs=10, n_part_tot=n_part_tot, n_threads=1, uncertainty_type="uncertainty"
    )
    print("------- test uncertainty for multiple runs -------")
    ok = (
        utility.assert_images_ratio_per_voxel(
            1, edep_test, edep_ref, abs_tolerance=0.03, mhd_is_path=False
        )
        and ok
    )
    ok = (
        utility.assert_images_ratio_per_voxel(
            1, unc_test, unc_ref, abs_tolerance=0.03, mhd_is_path=False
        )
        and ok
    )

    # STE of mean
    edep_ref, unc_ref = run_simulation(
        n_runs=1, n_part_tot=n_part_tot, n_threads=10, uncertainty_type="ste_of_mean"
    )
    edep_test, unc_test = run_simulation(
        n_runs=10, n_part_tot=n_part_tot, n_threads=10, uncertainty_type="ste_of_mean"
    )
    print("------- test ste_of_mean for multiple runs -------")
    ok = (
        utility.assert_images_ratio_per_voxel(
            1, edep_test, edep_ref, abs_tolerance=0.03, mhd_is_path=False
        )
        and ok
    )
    ok = (
        utility.assert_images_ratio_per_voxel(
            1, unc_test, unc_ref, abs_tolerance=0.03, mhd_is_path=False
        )
        and ok
    )

    utility.test_ok(ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.userhooks import check_production_cuts
from opengate.tests.utility import (
    get_default_test_paths,
    read_stat_file,
    assert_stats,
    test_ok,
)
from opengate.utility import g4_units
from opengate.logger import DEBUG, RUN
from opengate.managers import Simulation

if __name__ == "__main__":
    """
    The following line is only used for tests, it store the paths where the
    reference data are stored.
    """
    paths = get_default_test_paths(__file__, "gate_test004_simulation_stats_actor")

    """
    Create a simulation object. The class is 'gate.Simulation'.
    The single object that will contain all parameters of the
    simulation is called 'sim' here.
    """
    sim = Simulation()

    """
    Main global options.
    The 'sim' object contains a structure called 'user_info' that gather all global options.
    - For example here, the verbosity is set (verbosity means texts that are displayed during
    the simulation run, mostly for debug)
    - 'visu', if ON, display a windows with a QT view of the scene.
    - random_engine and random_seed control the pseudo random engine. We recommend MersenneTwister.
      A seed can be specified, e.g. 123456, for reproducible simulation. Or you can use 'auto', an random seed
      will be generated.
    """
    sim.verbose_level = DEBUG
    sim.running_verbose_level = RUN
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = True
    sim.visu_type = "vrml"
    sim.random_engine = "MersenneTwister"
    sim.random_seed = "auto"
    print(sim)

    """
    Units. Get some default units from G4. To define a value with a unit, e.g. do:
    x = 123 * cm
    """
    m = g4_units.m
    cm = g4_units.cm
    keV = g4_units.keV
    mm = g4_units.mm
    um = g4_units.um
    Bq = g4_units.Bq

    """
    Set the world size (like in the Gate macro). World is the only volume created by default.
    It is described by a dict-like structure, accessible by sim.world.
    The size is set here, as a 3D vector. Default material is G4_AIR.
    """
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    """
    A simple waterbox volume is created. It is inserted into the simulation with 'add_volume'.
    This function return a dict-like structure (called 'waterbox' here) with various parameters
    (size, position in the world, material). Note that, like in Geant4, the coordinate system
    of all volumes is the one of the mother volume (here the world).
    """
    waterbox = sim.add_volume("Box", "Waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    """
    The physic list by default is 'QGSP_BERT_EMV' (see Geant4 doc).
    """
    sim.physics_manager.physics_list_name = "QGSP_BERT_EMV"
    global_cut = 700 * um
    sim.physics_manager.global_production_cuts.gamma = global_cut
    sim.physics_manager.global_production_cuts.electron = global_cut
    sim.physics_manager.global_production_cuts.positron = global_cut
    sim.physics_manager.global_production_cuts.proton = global_cut

    """
    Create a source, called 'Default'. The type of the source is 'Generic'.
    Several parameters (particle, energy, direction etc) are available in the
    dict-like structure.
    """
    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 200000

    """
    Add a single scorer (called 'actor'), of type 'SimulationStatisticsActor'.
    This simple scorer store the number or Run/Events/Track/Steps of the simulation.
    We recommend to always add such actor.
    The flag 'track_types_flag' gives more detailed results about the tracks (particle type)
    """ ""
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    """
    Start the simulation ! You can relax and drink coffee.
    (The commented line indicates how to indicate to Geant4 to verbose during the simulation).
    """
    # sim.g4_commands_after_init.append("/run/verbose 1")
    sim.user_hook_after_init = check_production_cuts

    sim.visualize_geometry()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import uproot

# Create the simulation
import numpy as np
from opengate.tests import utility


def check_stats(path, fwhmX, fwhmY, fwhmZ):

    is_ok = True

    inFile = uproot.open(path)
    Singles_TG = inFile["Singles_crystal"]
    Singles_SB_TG = inFile["Singles_crystal_2"]

    # Convert branches to numpy arrays (fast and efficient)
    X_TG = Singles_TG["PostPosition_X"]
    Y_TG = Singles_TG["PostPosition_Y"]
    Z_TG = Singles_TG["PostPosition_Z"]

    X_SB_TG = Singles_SB_TG["PostPosition_X"]
    Y_SB_TG = Singles_SB_TG["PostPosition_Y"]
    Z_SB_TG = Singles_SB_TG["PostPosition_Z"]

    dX = np.asarray(X_TG) - np.asarray(X_SB_TG)
    dY = np.asarray(Y_TG) - np.asarray(Y_SB_TG)
    dZ = np.asarray(Z_TG) - np.asarray(Z_SB_TG)

    if (np.abs(np.std(dX) * 2.35 - fwhmX) / fwhmX) > 0.05:
        print("ERROR IN X!")
        is_ok = False

    if (np.abs(np.std(dY) * 2.35 - fwhmY) / fwhmY) > 0.05:
        print("ERROR IN Y!")
        is_ok = False

    if (np.abs(np.std(dZ) * 2.35 - fwhmZ) / fwhmZ) > 0.05:
        print("ERROR IN Z!")
        is_ok = False

    return is_ok


def main():
    paths = utility.get_default_test_paths(__file__, "", output_folder="test060")

    sim = gate.Simulation()
    sim.g4_verbose = False
    sim.visu = False
    sim.check_volumes_overlap = True
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"

    output_path = paths.output
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    KeV = gate.g4_units.keV
    kBq = gate.g4_units.Bq * 1000
    gcm3 = gate.g4_units.g_cm3
    sec = gate.g4_units.s

    sim.volume_manager.material_database.add_material_weights(
        "LYSO",
        ["Lu", "Y", "Si", "O"],
        [0.31101534, 0.368765605, 0.083209699, 0.237009356],
        5.37 * gcm3,
    )

    ###Test for GATE10 truncated Gaussian###

    # world
    world = sim.world
    world.size = [100 * cm, 100 * cm, 100 * cm]
    world.material = "G4_AIR"

    crystal = sim.add_volume("Box", name="crystal")
    crystal.size = [10 * cm, 10 * cm, 10 * cm]
    crystal.material = "LYSO"
    crystal.mother = world.name

    source = sim.add_source("GenericSource", "mysource")
    source.particle = "gamma"

    source.position.type = "box"
    source.position.size = [crystal.size[0] * 0.99, crystal.size[1] * 0.99, 0.1 * mm]
    source.position.translation = [0 * cm, 0 * cm, -5 * cm]

    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.energy.type = "mono"
    source.energy.mono = 511 * KeV
    source.activity = 20 * kBq

    # Hits
    hc = sim.add_actor("DigitizerHitsCollectionActor", f"Hits_{crystal.name}")
    hc.attached_to = crystal.name
    hc.output_filename = output_path / "output_singles_TEST.root"
    # hc.output = "output_config1.root"
    hc.attributes = [
        "EventID",
        "PostPosition",
        "PrePosition",
        "Position",
        "TotalEnergyDeposit",
        "PreStepUniqueVolumeID",
        "PostStepUniqueVolumeID",
        "GlobalTime",
    ]

    # Singles
    sc = sim.add_actor("DigitizerAdderActor", f"Singles_{crystal.name}")
    sc.attached_to = hc.attached_to
    sc.authorize_repeated_volumes = True
    sc.input_digi_collection = hc.name
    sc.policy = "EnergyWeightedCentroidPosition"
    sc.output_filename = hc.output_filename

    bc = sim.add_actor("DigitizerSpatialBlurringActor", f"Singles_{crystal.name}_2")
    bc.attached_to = hc.attached_to
    bc.output_filename = hc.output_filename
    bc.input_digi_collection = sc.name
    bc.keep_in_solid_limits = True
    bc.use_truncated_Gaussian = True
    bc.blur_attribute = "PostPosition"
    bc.blur_fwhm = [35 * mm, 20 * mm, 1 * mm]

    # timing
    sim.run_timing_intervals = [[0, 1 * sec]]
    # sim.run_timing_intervals = [[0, 0.01 * sec]]
    # go
    sim.run()

    is_ok = check_stats(hc.output_filename, *bc.blur_fwhm)
    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test038_gan_phsp_spect_gan_helpers as t38
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test038_gan_phsp_spect", "test038"
    )

    # create the simulation
    sim = gate.Simulation()
    sim.progress_bar = True
    condition_generator = t38.create_simulation(sim, paths, None)

    # add AA
    gsource = sim.get_source_user_info("gaga")
    gsource.skip_policy = "SkipEvents"
    gsource.direction.acceptance_angle.volumes = ["spect1"]
    gsource.direction.acceptance_angle.intersection_flag = True
    gsource.direction.acceptance_angle.normal_flag = False
    gsource.direction.acceptance_angle.max_rejection = 20000
    gsource.direction.acceptance_angle.skip_policy = "SkipEvents"
    # gsource.batch_size = 2e4

    # change output names
    stat = sim.actor_manager.get_actor("Stats")
    stat.output_filename = "test038_gan_aa_stats.txt"
    proj = sim.actor_manager.get_actor("Projection_spect1_crystal")
    proj.output_filename = "test038_gan_aa_proj.mhd"
    singles = sim.actor_manager.get_actor("Singles_spect1_crystal")
    singles.output_filename = "test038_gan_aa_singles.root"

    # go
    sim.run(start_new_process=True)

    #
    print()
    s = sim.source_manager.get_source("gaga")
    ref_se = 220534
    t_se = (ref_se - s.total_skipped_events) / ref_se * 100
    tol = 10
    is_ok = t_se < tol
    utility.print_test(
        is_ok,
        f"Source, nb of skipped particles (absorbed) : {s.total_skipped_events} {t_se:0.2f}% (tol = {tol}, {ref_se})",
    )

    print(
        f"Source, nb of zeros particles (absorbed) : {s.total_zero_events} (should be around 5)"
    )

    stats = sim.get_actor("Stats")
    stats_ref = utility.read_stat_file(paths.output_ref / "test038_gan_aa_stats.txt")
    # do not compare steps
    stats_ref.counts.steps = stats.counts.steps
    is_ok = utility.assert_stats(stats, stats_ref, 0.02) and is_ok

    stats.counts.events += s.total_skipped_events
    print("Number of events is increased by the nb of skipped events")
    print(stats)

    # image
    is_ok = (
        utility.assert_images(
            paths.output_ref / "test038_gan_aa_proj.mhd",
            paths.output / "test038_gan_aa_proj_counts.mhd",
            tolerance=70,
            axis="x",
            sum_tolerance=2.75,
            ignore_value_data2=0,
            apply_ignore_mask_to_sum_check=False,  # reproduce legacy behavior
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

# Physics list default range cut for protons, e+, e-, gamma
# defined in GEANT4/source/run/src/G4VUserPhysicsList.cc
DEFAULT_CUT = 1.0


def simulate(number_of_threads=1, start_new_process=False):
    # create the simulation
    energy = 200.0
    sim = gate.Simulation()
    sim.number_of_threads = number_of_threads

    # main options
    sim.g4_verbose = True
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_engine = "MersenneTwister"

    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV

    sim.physics_manager.physics_list_name = "G4EmStandardPhysics"
    global_cut = 1.34 * mm
    sim.physics_manager.global_production_cuts.electron = global_cut

    requested_cuts_proton = {}

    # *** Production cuts in a single volume ***
    waterbox_A = sim.add_volume("Box", "waterbox_A")
    waterbox_A.size = [10 * cm, 10 * cm, 10 * cm]
    waterbox_A.translation = [0 * cm, 0 * cm, 11 * cm]
    waterbox_A.material = "G4_WATER"

    # Choose an "awkward" cut
    # which does not corrispond to any of Geant4's
    # defaults to make the assertion (below) significant
    cut_proton = 10.7 * mm
    sim.physics_manager.set_production_cut(waterbox_A.name, "proton", cut_proton)
    requested_cuts_proton[waterbox_A.name] = cut_proton

    # *** Production cuts in individual volumes in nested volume structure ***
    waterbox_B = sim.add_volume("Box", "waterbox_B")
    waterbox_B.size = waterbox_A.size
    waterbox_B.translation = [
        0 * cm,
        0 * cm,
        waterbox_A.translation[2] + 1.1 * waterbox_B.size[2],
    ]
    waterbox_B.material = "G4_WATER"

    previous_mother = waterbox_B
    for i in range(6):
        new_insert = sim.add_volume("Box", f"insert_B_{i}")
        new_insert.size = [0.9 * s for s in previous_mother.size]
        new_insert.material = waterbox_B.material
        new_insert.mother = previous_mother.name
        previous_mother = new_insert
        # Set cut in every second insert
        if i % 2 == 0:
            cut_proton = 2.1 + i / 100.0 * mm
            sim.physics_manager.set_production_cut(
                new_insert.name, "proton", cut_proton
            )
            requested_cuts_proton[new_insert.name] = cut_proton

    # *** Production cuts propagated to nested volumes ***
    waterbox_C = sim.add_volume("Box", "waterbox_C")
    waterbox_C.size = waterbox_A.size
    waterbox_C.translation = [
        0 * cm,
        0 * cm,
        waterbox_B.translation[2] + 1.1 * waterbox_C.size[2],
    ]
    waterbox_C.material = "G4_WATER"

    cut_proton = 3.39 * mm
    sim.physics_manager.set_production_cut(waterbox_C.name, "proton", cut_proton)
    requested_cuts_proton[waterbox_C.name] = cut_proton

    previous_mother = waterbox_C
    for i in range(6):
        new_insert = sim.add_volume("Box", f"insert_C_{i}")
        new_insert.size = [0.9 * s for s in previous_mother.size]
        new_insert.material = waterbox_C.material
        new_insert.mother = previous_mother.name
        previous_mother = new_insert
        requested_cuts_proton[new_insert.name] = cut_proton

    # *** Production cuts set via region object ***
    region_D = sim.physics_manager.add_region("region_D")
    region_D.production_cuts.proton = 4.87 * mm

    for i in range(4):
        new_box = sim.add_volume("Box", f"waterbox_D{i}")
        new_box.size = [1 * mm, 1 * mm, 1 * mm]
        new_box.translation = [
            i * 2 * cm,
            0 * cm,
            waterbox_C.translation[2] + 1.1 * waterbox_C.size[2],
        ]
        new_box.material = "G4_WATER"
        region_D.associate_volume(new_box)
        requested_cuts_proton[new_box.name] = region_D.production_cuts.proton

    # Arbritrary source because we do not really need
    # the simulation, only the initialization
    source = sim.add_source("GenericSource", "Default")
    source.particle = "proton"
    source.energy.mono = energy * MeV
    source.position.radius = 1 * mm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 1e1

    print(sim.physics_manager.dump_production_cuts())

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # Set the hook function user_fct_after_init
    # to the function defined below
    sim.user_hook_after_init = check_production_cuts
    sim.run(start_new_process=start_new_process)

    # get results
    print(stats)
    print("track type", stats.counts.track_types)

    print("Checking production cuts:")
    retrieved_global_cut_proton = None
    for item in sim.user_hook_log:
        if item[0] == "world":
            retrieved_global_cut_proton = item[1]["proton"]
            retrieved_global_cut_positron = item[1]["positron"]
            retrieved_global_cut_electron = item[1]["electron"]
            retrieved_global_cut_gamma = item[1]["gamma"]

            print("retrieved_global_cut_positron")
            print(retrieved_global_cut_positron)

            # Verify that cuts are global cuts in volumes beloging to DefaultWorldRegion
            # global cut has been set for electron
            assert retrieved_global_cut_electron == global_cut
            # but not for the others, so the physics list default is applied
            assert retrieved_global_cut_proton == DEFAULT_CUT
            assert retrieved_global_cut_positron == DEFAULT_CUT
            assert retrieved_global_cut_gamma == DEFAULT_CUT

    for item in sim.user_hook_log:
        print(f"Volume {item[0]}:")
        value_dict = item[1]
        if item[0] != "world":
            if item[0] in requested_cuts_proton:
                print(
                    f"Requested production cut for protons: {requested_cuts_proton[item[0]]}"
                )
                print(f"Found: {value_dict['proton']}")
                assert requested_cuts_proton[item[0]] == value_dict["proton"]
            else:
                print(
                    f"Found production cut for protons, {value_dict['proton']}, but no requested cut."
                )
                if value_dict["proton"] == retrieved_global_cut_proton:
                    print(
                        "... but don't worry, this is just the global cut as expected."
                    )
                else:
                    print(
                        "... and that is strange because it does not match the global cut. "
                    )
                    # raise Exception("Found unexpected production cut")
            # NB: because region cuts were only set for protons
            # we should find the user-specified global cut for electrons
            assert value_dict["electron"] == global_cut
            # # and the physics list default (= 1.0) for the others
            assert value_dict["positron"] == DEFAULT_CUT
            assert value_dict["gamma"] == DEFAULT_CUT
            print(value_dict)

    print("Test passed")


def check_production_cuts(simulation_engine):
    """Function to be called by opengate after initialization
    of the simulation, i.e. when G4 volumes and regions exist.
    The purpose is to check whether Geant4 has properly set
    the production cuts in the specific region.

    The information is stored in the attribute hook_log
    which can be accessed via the output of the simulation.

    """
    print("Entered hook")
    for (
        volume_name,
        volume,
    ) in simulation_engine.simulation.volume_manager.volumes.items():
        # print(volume_name, g4_volume.g4_region)
        if volume.g4_region is not None:
            region_name = volume.g4_region.GetName()
            print(f"In hook: found volume {volume_name} with region {region_name}")
            user_limits = volume.g4_region.GetUserLimits()
            print(f"In hook: found UserLimit {user_limits}")

            pc = volume.g4_region.GetProductionCuts()
            cut_proton = pc.GetProductionCut("proton")
            # Note: G4 particle names != Gate names for e+ and e-
            cut_positron = pc.GetProductionCut("e+")
            cut_electron = pc.GetProductionCut("e-")
            cut_gamma = pc.GetProductionCut("gamma")
            simulation_engine.user_hook_log.append(
                (
                    volume_name,
                    {
                        "proton": cut_proton,
                        "positron": cut_positron,
                        "electron": cut_electron,
                        "gamma": cut_gamma,
                    },
                )
            )
            print(f"In hook: found production cut for protons {cut_proton}")


# --------------------------------------------------------------------------
if __name__ == "__main__":
    simulate(number_of_threads=1, start_new_process=False)
    # simulate(number_of_threads=2, start_new_process=False)
    # simulate(number_of_threads=1, start_new_process=True)
    # simulate(number_of_threads=2, start_new_process=True)
